import pytest
from os import urandom
from decimal import Decimal
import time
from pytaoassets.kutil import Kutil
from pytaoassets.provider import Explorer
from pytaoassets.networks import net_query
from pytaoassets.transactions import tx_output, find_parent_outputs, p2pkh_script
from taopy.structs.transaction import PeercoinMutableTx, PeercoinTx, Locktime
from taopy.structs.sig import P2pkhSolver


def test_key_generation():
    '''test privkey/pubkey generation.'''

    mykey = Kutil(network="xto")

    assert isinstance(mykey.privkey, str)
    assert isinstance(mykey.pubkey, str)


def test_key_generation_from_seed():
    '''check if key generation is what is expected from seed.'''

    seed = "Hello PeerAssets."
    mykey = Kutil(from_string=seed, network="xto")

    assert mykey.privkey == '680510f7f5e622347bc8d9e54e109a9192353693ef61d82d2d5bdf4bc9fd638b'
    assert mykey.pubkey == '037cf9e7664b5d10ce209cf9e2c7f68baa06f1950114f25677531b959edd7e670c'


def test_address_generation():
    '''test if addresses are properly made'''

    privkey = bytearray(urandom(32))

    assert Kutil(network="xto", privkey=privkey).address.startswith("P")

    assert isinstance(Kutil(network='xto').address, str)
    assert len(Kutil(network='xto').address) == 34


def test_mainnet_wif_import():
    '''test importing WIF privkey'''

    mykey = Kutil(network='xto', from_wif="CHTwXavkWCVi5PVFUDmCvRiJPZpzo1imLC9FEG4u4LakbnU7TU7f")

    assert mykey.address == 'TbdCceyr7sTxTKydxtDTPwoFhdDwY8T3Tv'
    assert mykey.pubkey == '0374113EA30D602A111DAFBECF0F7F5B72E139D257BB8ED7CC0E458E9594892C47'
    assert mykey.privkey == '4784326D5470603F569CF9BA257B7BAFD2A5D296AC61CEC44576FE7AD5EE0C7E'


def test_wif_export():
    '''test Kutil WIF export'''

    mykey = Kutil(network='xto',
                  privkey=bytearray.fromhex('1b19749afd007bf6db0029e0273a46409bc160b9349031752bbc3cd913bbbdd3')
                 )

    assert isinstance(mykey.wif, str)
    assert mykey.wif == 'CHTwXavkWCVi5PVFUDmCvRiJPZpzo1imLC9FEG4u4LakbnU7TU7f'


def test_sign_transaction():

    network_params = net_query('xto')
    provider = Explorer(network='xto')
    key = Kutil(network='xto',
                privkey=bytearray.fromhex('4784326D5470603F569CF9BA257B7BAFD2A5D296AC61CEC44576FE7AD5EE0C7E')
                )
    dest_address = 'mwn75Gavp6Y1tJxca53HeCj5zzERqWagr6'

    unspent = provider.select_inputs(key.address, 0.63)  # 0.69
    output = tx_output(network='xto',
                       value=Decimal(0.1),
                       n=0, script=p2pkh_script(network='xto',
                                                address=dest_address)
                       )

    unsigned = TaoMutableTx(version=1,
                                 timestamp=int(time.time()),
                                 ins=unspent['utxos'],
                                 outs=[output],
                                 network=network_params,
                                 locktime=Locktime(0)
                                 )

    parent_outputs = [find_parent_outputs(provider, i) for i in unsigned.ins]
    solver = P2pkhSolver(key._private_key)

    signed = unsigned.spend(parent_outputs,
                            [solver for i in parent_outputs])

    assert isinstance(signed, PeercoinTx)




'''Communicate with local or remote peercoin-daemon via JSON-RPC'''

from operator import itemgetter
from .common import Provider
from pytaoassets.exceptions import InsufficientFunds
from taopy.structs.transaction import MutableTxIn, Sequence, ScriptSig
from decimal import Decimal, getcontext
getcontext().prec = 6

try:
    #from peercoin_rpc import Client
    from taorpc.authproxy import AuthServiceProxy, JSONRPCException
except:
    raise EnvironmentError("taorpc library is required for this to work,\
                            use pip to install it.")

class Client:
    '''JSON-RPC Client.'''

    def __init__(self, testnet=False, username=None, password=None,
                 ip=None, port=None, directory=None):

        if not ip:
            self.ip = 'localhost'  # default to localhost
        else:
            self.ip = ip

        if not username and not password:
            if not directory:
                try:
                    self.username, self.password = self.userpass()  # try to read from ~/.ppcoin
                except:
                    self.username, self.password = self.userpass(dir='Tao')  # try to read from ~/.peercoin
            else:
                self.username, self.password = self.userpass(dir=directory)  # try some other directory

        else:
            self.username = username
            self.password = password
        if testnet is True:
            self.testnet = True
            self.port = 9904
            self.url = 'http://{0}:{1}'.format(self.ip, self.port)
        else:
            self.testnet = False
            self.port = 15151
            self.url = 'http://{0}:{1}'.format(self.ip, self.port)
        if port is not None:
            self.port = port
            self.url = 'http://{0}:{1}'.format(self.ip, self.port)

        self.session = requests.Session()
        self.session.auth = (self.username, self.password)
        self.session.headers.update({'content-type': 'application/json'})

    def userpass(self, dir='Tao'):
        '''Reads config file for username/password'''

        source = os.path.expanduser('~/.{0}/{0}.conf').format(dir)
        dest = open(source, 'r')
        with dest as conf:
            for line in conf:
                if line.startswith('rpcuser'):
                    username = line.split("=")[1].strip()
                if line.startswith("rpcpassword"):
                    password = line.split("=")[1].strip()

        return username, password

    def req(self, method, params=()):
        """send request to ppcoind"""

        response = self.session.post(self.url,
                   data=json.dumps({"method": method,
                                    "params": params,
                                    "jsonrpc": "1.1"})
                ).json()

        if response["error"] is not None:
            return response["error"]
        else:
            return response["result"]

    def batch(self, reqs: list ):
        """ send batch request using jsonrpc 2.0 """

        batch_data = []

        for req_id, req in enumerate(reqs):
            batch_data.append( {"method": req[0], "params": req[1], "jsonrpc": "2.0", "id": req_id} )

        data = json.dumps(batch_data)
        response = self.session.post(self.url, data=data).json()
        return response

class RpcNode(Client, Provider):
    '''JSON-RPC connection to local Peercoin node'''

    def select_inputs(self, address: str, amount: int) -> dict:
        '''finds apropriate utxo's to include in rawtx, while being careful
        to never spend old transactions with a lot of coin age.
        Argument is intiger, returns list of apropriate UTXO's'''

        utxos = []
        utxo_sum = Decimal(-0.01)  # starts from negative due to minimal fee
        for tx in sorted(self.listunspent(address=address), key=itemgetter('confirmations')):

            if tx["address"] not in (self.pa_parameters.P2TH_addr,
                                     self.pa_parameters.test_P2TH_addr):

                utxos.append(
                        MutableTxIn(txid=tx['txid'],
                                    txout=tx['vout'],
                                    sequence=Sequence.max(),
                                    script_sig=ScriptSig.empty())
                         )

                utxo_sum += Decimal(tx["amount"])
                if utxo_sum >= amount:
                    return {'utxos': utxos, 'total': utxo_sum}

        if utxo_sum < amount:
            raise InsufficientFunds("Insufficient funds.")

        raise Exception("undefined behavior :.(")

    @property
    def is_testnet(self) -> bool:
        '''check if node is configured to use testnet or mainnet'''

        if self.getinfo()["testnet"] is True:
            return True
        else:
            return False

    @property
    def network(self) -> str:
        '''return which network is the node operating on.'''

        if self.is_testnet:
            return "xto"
        else:
            return "xto"

    def listunspent(
        self,
        address: str="",
        minconf: int=1,
        maxconf: int=999999,
    ) -> list:
        '''list UTXOs
        modified version to allow filtering by address.
        '''
        if address:
            return self.req("listunspent", [minconf, maxconf, [address]])

        return self.req("listunspent", [minconf, maxconf])



from decimal import Decimal

from pytaoassets.provider.cryptoid import Cryptoid


def test_cryptoid_is_testnet():

    cryptoid = Cryptoid(network="ppc")

    assert isinstance(cryptoid.is_testnet, bool)
    assert cryptoid.is_testnet is False


def test_crypotid_network():

    assert Cryptoid(network="ppc").network == "peercoin"


def test_cryptoid_getblockcount():

    assert isinstance(Cryptoid(network="ppc").getblockcount(), int)


def test_cryptoid_getblock():

    provider = Cryptoid(network="tppc")
    assert isinstance(provider.getblock('0000000429a1e623da44a7430b9d9ae377bc2da203043c444c313b2d4390eba2'), dict)


def test_cryptoid_get_block_hash():

    assert isinstance(Cryptoid(network="ppc").getblockhash(3378), str)


def test_cryptoid_getdifficulty():

    difficulty = Cryptoid(network="ppc").getdifficulty()
    assert isinstance(difficulty["proof-of-stake"], float)


def test_cryptoid_getbalance():

    assert isinstance(Cryptoid(network="ppc").getbalance(
                      'PHvDhfz1dGyPbZZ3Qnp56y92zmy98sncZT'), Decimal)


def test_cryptoid_getreceivedbyaddress():

    assert isinstance(Cryptoid(network="ppc").getreceivedbyaddress(
                      'PHvDhfz1dGyPbZZ3Qnp56y92zmy98sncZT'), Decimal)


def test_cryptoid_listunspent():

    assert isinstance(Cryptoid(network="ppc").listunspent(
                      'PAdonateFczhZuKLkKHozrcyMJW7Y6TKvw'), list)


def test_cryptoid_getrawtransaction():

    assert isinstance(Cryptoid(network="ppc").getrawtransaction(
                      '34d19bf5a5c757d5bcbf83a91ad9bc04365c58a035a6bf728bce8013ad04c173', 1), dict)


def test_cryptoid_listtransactions():

    assert isinstance(Cryptoid(network="tppc").listtransactions(
                      'msPLoMcHpn6Y28pPKwATG411m5X7Vodu3m'), list)


def test_cryptoid_select_inputs():

    address = 'mvoSaPN8yTYWW7Tv3fVYTQnhkBuBWxpSP4'
    provider = Cryptoid(network='tppc')

    inputs = provider.select_inputs(address, 0.02)['utxos'][0]

    assert inputs.to_json() == {'scriptSig': {'asm': 'OP_DUP OP_HASH160 a7a826829741268ae2dd45942371000ea95d8524 OP_EQUALVERIFY OP_CHECKSIG',
                                'hex': '76a914a7a826829741268ae2dd45942371000ea95d852488ac'},
                                'sequence': '4294967295',
                                'txid': 'c637b4deb1ef552be42fd45a7cf8e4df0c7eeb5c4546d4343035ecf9669115ce',
                                'vout': 0
                                }



'''parse cards according to deck issue mode'''

from typing import Callable, Optional


def none_parser(cards: list) -> Optional[list]:
    '''
    parser for NONE [0] issue mode
    No issuance allowed.
    '''

    return None


def custom_parser(cards: list, parser: Optional[Callable[[list], Optional[list]]]=None) -> Optional[list]:
    '''parser for CUSTOM [1] issue mode,
    please provide your custom parser as argument'''

    if not parser:
        return cards

    else:
        return parser(cards)


def once_parser(cards: list) -> Optional[list]:
    '''
    parser for ONCE [2] issue mode
    Only one issuance transaction from asset owner allowed.
    '''

    first_issue = next(i for i in cards if i.type == "CardIssue")  # find first CardIssue

    filtered = [i for i in cards if i.type == "CardIssue" if i != first_issue]  #  drop all other CardIssues

    return [i for i in cards if i not in filtered]


def multi_parser(cards: list) -> Optional[list]:
    '''parser for MULTI [4] issue mode'''

    return cards


def mono_parser(cards: list) -> Optional[list]:
    '''
    parser for MONO [8] issue mode
    MONO = 0x08; // All card transaction amounts are equal to 1
    '''
    from pytaoassets.pautils import exponent_to_amount, amount_to_exponent

    processed_cards = []
    decimals = cards[0].number_of_decimals

    for c in cards:
        c.amount = [amount_to_exponent(
                     exponent_to_amount(c.amount[0], decimals),
                     decimals)]
        processed_cards.append(c)

    return processed_cards


def unflushable_parser(cards: list) -> Optional[list]:
    '''
    parser for UNFLUSHABLE [16] issue mode
    No card transfer transactions allowed except for the card-issue transaction
    '''

    return [i for i in cards if i.type == "CardIssue"]


parsers = {
    'NONE': none_parser,
    'CUSTOM': custom_parser,
    'ONCE': once_parser,
    'MULTI': multi_parser,
    'MONO': mono_parser,
    'UNFLUSHABLE': unflushable_parser
}



import pytest

from pytaoassets.exceptions import UnsupportedNetwork
from pytaoassets.pa_constants import param_query


def test_param_query():
    "Check that we can find PAParams for networks by name."

    # Use a network's long name
    pa_params = param_query("peercoin")
    assert pa_params.network_shortname == "ppc"

    # Use a network's short name
    pa_params = param_query("tppc")
    assert pa_params.network_name == "peercoin-testnet"

    # Try to find a network we don't know about.
    with pytest.raises(UnsupportedNetwork):
        param_query("not a network name we know")




from hashlib import sha256
from typing import Union
from os import urandom

from taopy.structs.crypto import PublicKey, PrivateKey
from taopy.structs.transaction import MutableTransaction, TxOut
from taopy.structs.sig import P2pkhSolver

from pytaoassets.networks import net_query


class Kutil:

    def __init__(self, network: str, privkey: bytearray=None, from_string: str=None,
                 from_wif: str=None) -> None:
        '''
           High level helper class for handling public key cryptography.

           : privkey - privatekey bytes
           : from_wif - <WIF> import private key from your wallet in WIF format
           : from_bytes - import private key in binary format
           : network - specify network [ppc, tppc, btc]
           : from_string - specify seed (string) to make the privkey from
           '''

        self.network = network
        self.constants = net_query(self.network)

        if privkey is not None:
            self._private_key = PrivateKey(privkey)

        if from_string is not None:
            self._private_key = PrivateKey(sha256(
                                           from_string.encode()).digest())

        if from_wif is not None:
            self._private_key = PrivateKey.from_wif(wif=from_wif,
                                                    network=self.constants,
                                                    )

        if not privkey:
            if from_string == from_wif is None:  # generate a new privkey
                self._private_key = PrivateKey(bytearray(urandom(32)))

        self.privkey = str(self._private_key)
        self._public_key = PublicKey.from_priv(self._private_key)
        self.pubkey = str(self._public_key)

    @property
    def address(self) -> str:
        '''generate an address from pubkey'''

        return str(self._public_key.to_address(
                   net_query(self.network))
                   )

    @property
    def wif(self) -> str:
        '''convert raw private key to WIF'''

        return self._private_key.to_wif(network=self.constants)

    def sign_transaction(self, txins: Union[TxOut],
                         tx: MutableTransaction) -> MutableTransaction:
        '''sign the parent txn outputs P2PKH'''

        solver = P2pkhSolver(self._private_key)
        return tx.spend(txins, [solver for i in txins])



from collections import namedtuple
from decimal import Decimal
from pytaoassets.exceptions import UnsupportedNetwork

# constants to be consumed by the backend
Constants = namedtuple('Constants', [
    'name',
    'shortname',
    'base58_prefixes',
    'base58_raw_prefixes',
    'bech32_hrp',
    'bech32_net',
    'xkeys_prefix',
    'xpub_version',
    'xprv_version',
    'wif_prefix',
    'from_unit',
    'to_unit',
    'min_tx_fee',
    'tx_timestamp',
    'op_return_max_bytes'
])


'''
Network name should be lowercase, for testnet append "-testnet".
For abbreviation prefix testnet of the network with "t".
'''

networks = (

    # Tao mainnet
    Constants(
        name='tao',
        shortname='xto',
        base58_prefixes={
            'T': 'p2pkh',
            '2': 'p2sh',
        },
        base58_raw_prefixes={
            'p2pkh': bytearray(b'\x42'),
            'p2sh': bytearray(b'\x03'),
        },
        bech32_hrp='tc',
        bech32_net='mainnet',
        xkeys_prefix='x',
        xpub_version=b'\x04\x88\xb2\x1e',
        xprv_version=b'\x04\x88\xad\xe4',
        wif_prefix=0x4c,
        from_unit=Decimal('1e-6'),
        to_unit=Decimal('1e6'),
        min_tx_fee=Decimal(0.0002),
        tx_timestamp=True,
        op_return_max_bytes=80
    ),

    Constants(
        name='peercoin',
        shortname='ppc',
        base58_prefixes={
            'P': 'p2pkh',
            'p': 'p2sh',
        },
        base58_raw_prefixes={
            'p2pkh': bytearray(b'\x37'),
            'p2sh': bytearray(b'\x75'),
        },
        bech32_hrp='bc',
        bech32_net='mainnet',
        xkeys_prefix='x',
        xpub_version=b'\x04\x88\xb2\x1e',
        xprv_version=b'\x04\x88\xad\xe4',
        wif_prefix=0xb7,
        from_unit=Decimal('1e-6'),
        to_unit=Decimal('1e6'),
        min_tx_fee=Decimal(0.01),
        tx_timestamp=True,
        op_return_max_bytes=80
    ),

    # Peercoin testnet
    Constants(
        name='peercoin-testnet',
        shortname='tppc',
        base58_prefixes={
            'm': 'p2pkh',
            'n': 'p2pkh',
        },
        base58_raw_prefixes={
            'p2pkh': bytearray(b'\x6f'),
            'p2sh': bytearray(b'\xc4'),
        },
        bech32_hrp='tb',
        bech32_net='testnet',
        xkeys_prefix='t',
        xpub_version=b'\x04\x35\x87\xcf',
        xprv_version=b'\x04\x35\x83\x94',
        wif_prefix=0xef,
        from_unit=Decimal('1e-6'),
        to_unit=Decimal('1e6'),
        min_tx_fee=Decimal(0.01),
        tx_timestamp=True,
        op_return_max_bytes=80
    )
)


def net_query(name: str) -> Constants:
    '''Find the NetworkParams for a network by its long or short name. Raises
    UnsupportedNetwork if no NetworkParams is found.
    '''

    for net_params in networks:
        if name in (net_params.name, net_params.shortname,):
            return net_params

    raise UnsupportedNetwork



import pytest
from decimal import Decimal
from pytaoassets.provider.explorer import Explorer


def test_explorer_is_testnet():

    explorer = Explorer(network="tppc")

    assert isinstance(explorer.is_testnet, bool)
    assert explorer.is_testnet is True


def test_explorer_network():

    assert Explorer(network="ppc").network == "peercoin"


def test_explorer_getblockcount():

    assert isinstance(Explorer(network="ppc").getblockcount(), int)


def test_explorer_getblock():

    provider = Explorer(network="ppc")
    assert isinstance(provider.getblock('00000000000da9a26b4f4ce3f1f286438ec2198e5f60d108fafa700061b486e7'), dict)


def test_explorer_get_block_hash():

    assert isinstance(Explorer(network="ppc").getblockhash(3378), str)


def test_explorer_getdifficulty():

    difficulty = Explorer(network="ppc").getdifficulty()
    assert isinstance(difficulty["proof-of-stake"], float)
    assert isinstance(difficulty["proof-of-work"], float)


def test_explorer_getbalance():

    assert isinstance(Explorer(network="ppc").getbalance(
                      'PHvDhfz1dGyPbZZ3Qnp56y92zmy98sncZT'), Decimal)


def test_explorer_getreceivedbyaddress():

    assert isinstance(Explorer(network="ppc").getreceivedbyaddress(
                      'PHvDhfz1dGyPbZZ3Qnp56y92zmy98sncZT'), Decimal)


def test_explorer_listunspent():

    assert isinstance(Explorer(network="ppc").listunspent(
                      'PAdonateFczhZuKLkKHozrcyMJW7Y6TKvw'), list)


@pytest.mark.parametrize("decrypt", [0, 1])
def test_explorer_getrawtransaction(decrypt):

    if decrypt:
        assert isinstance(Explorer(network="ppc").getrawtransaction(
                          '34d19bf5a5c757d5bcbf83a91ad9bc04365c58a035a6bf728bce8013ad04c173', decrypt), dict)
    else:
        assert isinstance(Explorer(network="ppc").getrawtransaction(
                          '34d19bf5a5c757d5bcbf83a91ad9bc04365c58a035a6bf728bce8013ad04c173', decrypt), str)


def test_explorer_listtransactions():

    assert isinstance(Explorer(network="ppc").listtransactions(
                      'PHvDhfz1dGyPbZZ3Qnp56y92zmy98sncZT'), list)


def test_explorer_select_inputs():

    address = 'mvoSaPN8yTYWW7Tv3fVYTQnhkBuBWxpSP4'
    provider = Explorer(network='tppc')

    inputs = provider.select_inputs(address, 0.02)['utxos'][0]

    assert inputs.to_json() == {'scriptSig': {'asm': 'OP_DUP OP_HASH160 a7a826829741268ae2dd45942371000ea95d8524 OP_EQUALVERIFY OP_CHECKSIG',
                                'hex': '76a914a7a826829741268ae2dd45942371000ea95d852488ac'},
                                'sequence': '4294967295',
                                'txid': 'c637b4deb1ef552be42fd45a7cf8e4df0c7eeb5c4546d4343035ecf9669115ce',
                                'vout': 0
                                }



'''Common provider class with basic features.'''

from abc import ABC, abstractmethod
from decimal import Decimal
import urllib.request

from taopy.structs.address import Address

from pytaoassets.exceptions import UnsupportedNetwork
from pytaoassets.pa_constants import PAParams, param_query
from pytaoassets.networks import Constants, net_query


class Provider(ABC):

    net = ""

    headers = {"User-Agent": "pytaoassets"}

    @staticmethod
    def _netname(name: str) -> dict:
        '''resolute network name,
        required because some providers use shortnames and other use longnames.'''

        try:
            long = net_query(name).name
            short = net_query(name).shortname
        except AttributeError:
            raise UnsupportedNetwork('''This blockchain network is not supported by the pytaoassets, check networks.py for list of supported networks.''')

        return {'long': long,
                'short': short}

    @property
    def network(self) -> str:
        '''return network full name'''

        return self._netname(self.net)['long']

    @property
    def pa_parameters(self) -> PAParams:
        '''load network PeerAssets parameters.'''

        return param_query(self.network)

    @property
    def network_properties(self) -> Constants:
        '''network parameters [min_fee, denomination, ...]'''

        return net_query(self.network)

    @property
    def is_testnet(self) -> bool:
        """testnet or not?"""

        if "testnet" in self.network:
            return True
        else:
            return False

    @classmethod
    def sendrawtransaction(cls, rawtxn: str) -> str:
        '''sendrawtransaction remote API'''

        if cls.is_testnet:
            url = 'https://testnet-explorer.peercoin.net/api/sendrawtransaction?hex={0}'.format(rawtxn)
        else:
            url = 'https://explorer.peercoin.net/api/sendrawtransaction?hex={0}'.format(rawtxn)

        resp = urllib.request.urlopen(url)
        return resp.read().decode('utf-8')

    @abstractmethod
    def getblockhash(self, blocknum: int) -> str:
        '''get blockhash using blocknum query'''
        raise NotImplementedError

    @abstractmethod
    def getblockcount(self) -> int:
        '''get block count'''
        raise NotImplementedError

    @abstractmethod
    def getblock(self, hash: str) -> dict:
        '''query block using <blockhash> as key.'''
        raise NotImplementedError

    @abstractmethod
    def getdifficulty(self) -> dict:
        raise NotImplementedError

    @abstractmethod
    def getbalance(self, address: str) -> Decimal:
        raise NotImplementedError

    @abstractmethod
    def getreceivedbyaddress(self, address: str) -> Decimal:
        raise NotImplementedError

    @abstractmethod
    def listunspent(self, address: str) -> list:
        raise NotImplementedError

    @abstractmethod
    def select_inputs(self, address: str, amount: int) -> dict:
        raise NotImplementedError

    @abstractmethod
    def getrawtransaction(self, txid: str, decrypt: int=1) -> dict:
        raise NotImplementedError

    @abstractmethod
    def listtransactions(self, address: str) -> list:
        raise NotImplementedError

    def validateaddress(self, address: str) -> bool:
        "Returns True if the passed address is valid, False otherwise."

        try:
            Address.from_string(self.network_properties, address)
        except ValueError:
            return False

        return True




'''miscellaneous utilities.'''

from pytaoassets.provider import Provider, RpcNode, Explorer, Cryptoid

from pytaoassets.exceptions import (InvalidDeckSpawn,
                                     InvalidDeckMetainfo,
                                     InvalidDeckIssueMode,
                                     InvalidDeckVersion,
                                     InvalidNulldataOutput,
                                     DeckP2THImportError,
                                     P2THImportFailed)

from pytaoassets.exceptions import (InvalidCardTransferP2TH,
                                     CardVersionMismatch,
                                     CardNumberOfDecimalsMismatch,
                                     RecieverAmountMismatch
                                     )

from google.protobuf.message import DecodeError
from pytaoassets.pa_constants import param_query
from typing import Iterable, Iterator, Optional, Tuple, List

from pytaoassets.paproto_pb2 import DeckSpawn as DeckSpawnProto
from pytaoassets.paproto_pb2 import CardTransfer as CardTransferProto
from pytaoassets.protocol import Deck, CardTransfer, CardBundle

def load_p2th_privkey_into_local_node(provider: RpcNode, prod: bool=True) -> None:
    '''Load PeerAssets P2TH privkey into the local node.'''

    assert isinstance(provider, RpcNode), {"error": "Import only works with local node."}
    error = {"error": "Loading P2TH privkey failed."}
    pa_params = param_query(provider.network)

    if prod:
        provider.importprivkey(pa_params.P2TH_wif, "PAPROD")
        #  now verify if ismine == True
        if not provider.validateaddress(pa_params.P2TH_addr)['ismine']:
            raise P2THImportFailed(error)
    else:
        provider.importprivkey(pa_params.test_P2TH_wif, "PATEST")
        if not provider.validateaddress(pa_params.test_P2TH_addr)['ismine']:
            raise P2THImportFailed(error)


def find_tx_sender(provider: Provider, raw_tx: dict) -> str:
    '''find transaction sender, vin[0] is used in this case.'''

    vin = raw_tx["vin"][0]
    txid = vin["txid"]
    index = vin["vout"]
    return provider.getrawtransaction(txid, 1)["vout"][index]["scriptPubKey"]["addresses"][0]


def find_deck_spawns(provider: Provider, prod: bool=True) -> Iterable[str]:
    '''find deck spawn transactions via Provider,
    it requires that Deck spawn P2TH were imported in local node or
    that remote API knows about P2TH address.'''

    pa_params = param_query(provider.network)

    if isinstance(provider, RpcNode):

        if prod:
            decks = (i["txid"] for i in provider.listtransactions("PAPROD"))
        else:
            decks = (i["txid"] for i in provider.listtransactions("PATEST"))

    if isinstance(provider, Cryptoid) or isinstance(provider, Explorer):

        if prod:
            decks = (i for i in provider.listtransactions(pa_params.P2TH_addr))
        else:
            decks = (i for i in provider.listtransactions(pa_params.test_P2TH_addr))

    return decks


def deck_parser(args: Tuple[Provider, dict, int, str],
                prod: bool=True) -> Optional[Deck]:
    '''deck parser function'''

    provider = args[0]
    raw_tx = args[1]
    deck_version = args[2]
    p2th = args[3]

    try:
        validate_deckspawn_p2th(provider, raw_tx, p2th)

        d = parse_deckspawn_metainfo(read_tx_opreturn(raw_tx['vout'][1]),
                                     deck_version)

        if d:

            d["id"] = raw_tx["txid"]
            try:
                d["issue_time"] = raw_tx["blocktime"]
            except KeyError:
                d["time"] = 0
            d["issuer"] = find_tx_sender(provider, raw_tx)
            d["network"] = provider.network
            d["production"] = prod
            d["tx_confirmations"] = raw_tx["confirmations"]
            return Deck(**d)

    except (InvalidDeckSpawn, InvalidDeckMetainfo, InvalidDeckVersion,
            InvalidNulldataOutput) as err:
        pass

    return None


def tx_serialization_order(provider: Provider, blockhash: str, txid: str) -> int:
    '''find index of this tx in the blockid'''

    return provider.getblock(blockhash)["tx"].index(txid)


def read_tx_opreturn(vout: dict) -> bytes:
    '''Decode OP_RETURN message from vout[1]'''

    asm = vout['scriptPubKey']['asm']
    n = asm.find('OP_RETURN')
    if n == -1:
        raise InvalidNulldataOutput({'error': 'OP_RETURN not found.'})
    else:
        # add 10 because 'OP_RETURN ' is 10 characters
        n += 10
        data = asm[n:]
        n = data.find(' ')
        # make sure that we don't include trailing opcodes
        if n == -1:
            return bytes.fromhex(data)
        else:
            return bytes.fromhex(data[:n])


def deck_issue_mode(proto: DeckSpawnProto) -> Iterable[str]:
    '''interpret issue mode bitfeg'''

    if proto.issue_mode == 0:
        yield "NONE"
        return

    for mode, value in proto.MODE.items():
        if value > proto.issue_mode:
            continue
        if value & proto.issue_mode:
            yield mode


def issue_mode_to_enum(deck: DeckSpawnProto, issue_mode: list) -> int:
    '''encode issue mode(s) as bitfeg'''

    # case where there are multiple issue modes specified
    if isinstance(issue_mode, list) and len(issue_mode) > 1:
        r = 0
        for mode in issue_mode:
            r += deck.MODE.Value(mode)
        return r

    elif isinstance(issue_mode, str):  # if single issue mode
        return deck.MODE.Value(issue_mode)

    else:
        raise InvalidDeckIssueMode({'error': 'issue_mode given in wrong format.'})


def parse_deckspawn_metainfo(protobuf: bytes, version: int) -> dict:
    '''Decode deck_spawn tx op_return protobuf message and validate it,
       Raise error if deck_spawn metainfo incomplete or version mistmatch.'''

    deck = DeckSpawnProto()
    deck.ParseFromString(protobuf)

    error = {"error": "Deck ({deck}) metainfo incomplete, deck must have a name.".format(deck=deck.name)}

    if deck.name == "":
        raise InvalidDeckMetainfo(error)

    if deck.version != version:
        raise InvalidDeckVersion({"error", "Deck version mismatch."})

    return {
        "version": deck.version,
        "name": deck.name,
        "issue_mode": deck.issue_mode,
        "number_of_decimals": deck.number_of_decimals,
        "asset_specific_data": deck.asset_specific_data
        }


def validate_deckspawn_p2th(provider: Provider, rawtx: dict, p2th: str) -> bool:
    '''Return True if deck spawn pays to p2th in vout[0] and if the P2TH address
    is correct. Otherwise raises InvalidDeckSpawn.
    '''

    try:
        vout = rawtx["vout"][0]["scriptPubKey"].get("addresses")[0]
    except TypeError:
        '''TypeError: 'NoneType' object is not subscriptable error on some of the deck spawns.'''
        raise InvalidDeckSpawn("Invalid Deck P2TH.")

    if not vout == p2th:
        raise InvalidDeckSpawn("InvalidDeck P2TH.")

    return True


def load_deck_p2th_into_local_node(provider: RpcNode, deck: Deck) -> None:
    '''
    load deck p2th into local node via "importprivke",
    this allows building of proof-of-timeline for this deck
    '''

    assert isinstance(provider, RpcNode), {"error": "You can load privkeys only into local node."}
    error = {"error": "Deck P2TH import went wrong."}

    provider.importprivkey(deck.p2th_wif, deck.id)
    check_addr = provider.validateaddress(deck.p2th_address)

    if not check_addr["isvalid"] and not check_addr["ismine"]:
        raise DeckP2THImportError(error)


def validate_card_transfer_p2th(deck: Deck, vout: dict) -> None:
    '''validate if card_transfer transaction pays to deck p2th in vout[0]'''

    error = {"error": "Card transfer is not properly tagged."}

    try:
        address = vout["scriptPubKey"].get("addresses")[0]
        if not address == deck.p2th_address:
            raise InvalidCardTransferP2TH(error)
    except TypeError as e:
        raise e


def parse_card_transfer_metainfo(protobuf: bytes, deck_version: int) -> dict:
    '''decode card_spawn protobuf message and validate it against deck.version
    :protobuf - bytes from op_return message
    :deck_version - integer
    '''

    card = CardTransferProto()
    card.ParseFromString(protobuf)

    if not card.version == deck_version:
        raise CardVersionMismatch({'error': 'card version does not match deck version.'})

    return {
        "version": card.version,
        "number_of_decimals": card.number_of_decimals,
        "amount": list(card.amount),
        "asset_specific_data": card.asset_specific_data
    }


def card_postprocess(card: dict, vout: list) -> List[dict]:

    # if card states multiple outputs, interpert it as a batch
    if len(card["amount"]) > 1:
        cards = []
        for am, v in zip(card["amount"], vout[2:]):
            c = card.copy()
            c["amount"] = [am]
            c["receiver"] = v["scriptPubKey"]["addresses"]
            c["cardseq"] = vout[2:].index(v)

            cards.append(c)
        return cards

    else:
        card["receiver"] = vout[2]["scriptPubKey"]["addresses"]
        card["cardseq"] = 0

    return [card]


def bundle_parser(bundle: CardBundle) -> Iterator:
    '''this function wraps all the card transfer parsing'''

    try:
        # first vout of the bundle must pay to deck.p2th
        validate_card_transfer_p2th(bundle.deck, bundle.vouts[0])

        # second vout must be OP_RETURN with card_metainfo
        card_metainfo = parse_card_transfer_metainfo(
                            read_tx_opreturn(bundle.vouts[1]),
                            bundle.deck.version
                            )

    # if any of this exceptions is raised, return None
    except (InvalidCardTransferP2TH, CardVersionMismatch,
            CardNumberOfDecimalsMismatch,
            RecieverAmountMismatch, DecodeError, TypeError,
            InvalidNulldataOutput) as e:

        #print(e)  # re-do as logging later on
        return
        yield

    # check for decimals
    if not card_metainfo["number_of_decimals"] == bundle.deck.number_of_decimals:
        raise CardNumberOfDecimalsMismatch(
                {"error": "Number of decimals does not match."}
                )

    # deduce the individual cards in the bundle
    cards = card_postprocess(card_metainfo, bundle.vouts)

    #  drop the vouts property
    del bundle.__dict__['vouts']

    for c in cards:
        d = {**c, **bundle.__dict__}
        yield CardTransfer(**d)


def amount_to_exponent(amount: float, number_of_decimals: int) -> int:
    '''encode amount integer as exponent'''

    return int(amount * 10**number_of_decimals)


def exponent_to_amount(exponent: int, number_of_decimals: int) -> float:
    '''exponent to integer to be written on the chain'''

    return exponent / 10**number_of_decimals



'''all custom exceptions should go here'''


class EmptyP2THDirectory(Exception):
    '''no transactions on this P2TH directory'''


class P2THImportFailed(Exception):
    '''Importing of PeerAssets P2TH privkeys failed.'''


class InvalidDeckIssueModeCombo(Exception):
    '''When verfiying deck issue_mode combinations.'''


class UnsupportedNetwork(Exception):
    '''This network is not suppored by pytaoassets.'''


class InvalidDeckSpawn(Exception):
    '''Invalid deck_spawn, deck is not properly tagged.'''


class InvalidDeckMetainfo(Exception):
    '''Deck metainfo incomplete, deck must have a name.'''


class InvalidDeckVersion(Exception):
    '''Deck version mistmatch.'''


class InvalidDeckIssueMode(Exception):
    '''Deck Issue mode is wrong.'''


class DeckP2THImportError(Exception):
    '''When Deck P2TH import goes wrong.'''


class InvalidCardTransferP2TH(Exception):
    '''card_transfer does not pay to deck p2th in vout[0]'''


class CardVersionMismatch(Exception):
    '''card_transfers version must match deck.version'''


class CardNumberOfDecimalsMismatch(Exception):
    '''card_tranfer number of decimals does not match deck rules.'''


class InvalidCardIssue(Exception):
    '''deck issuer can't issue cards to the deck issuing address'''


class RecieverAmountMismatch(Exception):
    '''card_transfer list of recievers is not equal to list of amounts'''


class InsufficientFunds(Exception):
    '''this address does not have enough assigned UTXOs'''


class InvalidNulldataOutput(Exception):
    '''mallformed OP_RETURN transaction output.'''


class InvalidVoutOrder(Exception):
    '''mallformed vout sequence'''


class OverSizeOPReturn(Exception):
    '''op_return size is exceeding the maximum size allowed by this network.'''



from typing import Generator

import pytest

from pytaoassets import (
    Cryptoid,
    Deck,
    Explorer,
    RpcNode,
    find_deck,
)
from pytaoassets.exceptions import *
from pytaoassets.paproto_pb2 import DeckSpawn
from pytaoassets.pautils import *
from pytaoassets.protocol import IssueMode, CardTransfer
from pytaoassets.pa_constants import param_query


@pytest.mark.xfail
def test_load_p2th_privkeys_into_local_node():

    provider = RpcNode(testnet=True)
    load_p2th_privkeys_into_local_node(provider=provider)


@pytest.mark.parametrize("prov", ["explorer", "cryptoid"])
def test_find_tx_sender(prov):

    if prov == "explorer":
        provider = Explorer(network="peercoin")
        rawtx = provider.getrawtransaction("397bda2f5e6608c872a663b2e5482d95db8ecfad00757823f0f12caa45a213a6", 1)
        assert find_tx_sender(provider, rawtx) == 'PNHGzKupyvo2YZVb1CTdRxtCGBB5ykgiug'

    if prov == "cryptoid":
        provider = Cryptoid(network="peercoin")
        rawtx = provider.getrawtransaction("397bda2f5e6608c872a663b2e5482d95db8ecfad00757823f0f12caa45a213a6", 1)
        assert find_tx_sender(provider, rawtx) == 'PNHGzKupyvo2YZVb1CTdRxtCGBB5ykgiug'


@pytest.mark.parametrize("prov", ["explorer", "cryptoid"])
def test_find_deck_spawns(prov):

    if prov == "explorer":
        provider = Explorer(network="peercoin")

    if prov == "cryptoid":
        provider = Cryptoid(network="peercoin")

    assert isinstance(find_deck_spawns(provider), Generator)


@pytest.mark.parametrize("prov", ["rpc", "explorer"])
def test_tx_serialization_order(prov):

    if prov == "explorer":
        provider = Explorer(network="peercoin-testnet")
        assert tx_serialization_order(provider,
                                      txid="f968702bcedc107959aae2c2b1a1becdccbfe7e5a32b460b2c13c1adaa33d541", blockhash="e234d2ef69f7cd1e7ee489546b39314cc838763b4e32438106cba657d9749f2f") == 1

    try:
        if prov == "rpc":
            provider = RpcNode(testnet=True)
            assert tx_serialization_order(provider,
                                          txid="f968702bcedc107959aae2c2b1a1becdccbfe7e5a32b460b2c13c1adaa33d541", blockhash="e234d2ef69f7cd1e7ee489546b39314cc838763b4e32438106cba657d9749f2f") == 1

    except:
        print("No RpcNode avaliable.")


def test_read_tx_opreturn():

    vout = [{'n': 0,
             'scriptPubKey': {'addresses': ['miHhMLaMWubq4Wx6SdTEqZcUHEGp8RKMZt'],
             'asm': 'OP_DUP OP_HASH160 1e667ee94ea8e62c63fe59a0269bb3c091c86ca3 OP_EQUALVERIFY OP_CHECKSIG',
             'hex': '76a9141e667ee94ea8e62c63fe59a0269bb3c091c86ca388ac',
             'reqSigs': 1,
             'type': 'pubkeyhash'},
             'value': 0.01},
             {'n': 1,
              'scriptPubKey': {'asm': 'OP_RETURN 0801120f736978746f5f726f6472696775657a18052004',
              'hex': '6a170801120f736978746f5f726f6472696775657a18052004',
              'type': 'nulldata'},
              'value': 0}
             ]

    assert isinstance(read_tx_opreturn(vout[1]), bytes)
    assert read_tx_opreturn(vout[1]) == b'\x08\x01\x12\x0fsixto_rodriguez\x18\x05 \x04'


def generate_dummy_deck():

    return Deck(
        name="decky",
        number_of_decimals=2,
        issue_mode=IssueMode.SINGLET.value,
        network="ppc",
        production=True,
        version=1,
        asset_specific_data="just testing.",
    )


def test_deck_issue_mode():
    '''test enum to issue_mode conversion'''

    deck_meta = DeckSpawn()
    deck_meta.issue_mode = 3

    assert isinstance(deck_issue_mode(deck_meta), Generator)
    assert list(deck_issue_mode(deck_meta)) == ['CUSTOM', 'ONCE']

    # Check that we handle NONE mode correctly.
    deck_meta.issue_mode = 0
    assert list(deck_issue_mode(deck_meta)) == ['NONE']


def test_issue_mode_to_enum():
    '''test issue mode to enum conversion'''

    deck = generate_dummy_deck().metainfo_to_protobuf
    deck_meta = DeckSpawn()
    deck_meta.ParseFromString(deck)

    assert isinstance(issue_mode_to_enum(deck_meta,
                      ["CUSTOM", "SINGLET"]), int)


def test_parse_deckspawn_metainfo():
    '''tests if loading of deck parameteres from protobuf works as it should.'''

    string = b'\x08\x01\x12\x0cmy_test_deck\x18\x03 \x02'
    assert parse_deckspawn_metainfo(string, 1) == {'issue_mode': IssueMode.ONCE.value,
                                                   'name': 'my_test_deck',
                                                   'number_of_decimals': 3,
                                                   'version': 1,
                                                   'asset_specific_data': b''
                                                  }

    string = b'\x08\x01\x18\x05 \x04' # without deck name
    with pytest.raises(InvalidDeckMetainfo):
        parse_deckspawn_metainfo(string, 1)


def test_validate_deckspawn_p2th():
    '''test deckspawn p2th validation'''

    provider = Explorer(network="peercoin-testnet")
    p2th = param_query('peercoin-testnet').P2TH_addr
    raw_tx = provider.getrawtransaction('643dccd585211766fc03f71e92fbf299cfc2bdbf3f2cae0ad85adec3141069f3', 1,)

    assert validate_deckspawn_p2th(provider, raw_tx, p2th)


@pytest.mark.xfail
def test_load_deck_p2th_into_local_node():

    provider = RpcNode(testnet=True)
    deck = generate_dummy_deck()
    load_deck_p2th_into_local_node(provider, deck)


def test_validate_card_transfer_p2th():

    provider = Cryptoid(network="peercoin-testnet")
    deck = find_deck(provider, "643dccd585211766fc03f71e92fbf299cfc2bdbf3f2cae0ad85adec3141069f3", 1)
    raw_tx = provider.getrawtransaction("809c506bc3add9e46a4d3a65348426688545213da5fb5b524acd380f2cdaf3cc", 1)

    validate_card_transfer_p2th(deck, raw_tx['vout'][0])


def test_parse_card_transfer_metainfo():

    card = b'\x08\x01\x12\n\xd0\xd2=\x80\x89z\xee\x83\xb8\x01\x18\x05'
    res = parse_card_transfer_metainfo(card, 1)

    assert isinstance(res, dict)


def test_card_postprocess():

    card = {'amount': [1000000], 'number_of_decimals': 3, 'version': 1}
    vout = [
            {
                "value": 0.01,
                "n": 0,
                "scriptPubKey": {
                    "asm": "OP_DUP OP_HASH160 3d9df85b2c05f0c95347e1738034e0653cd61269 OP_EQUALVERIFY OP_CHECKSIG",
                    "hex": "76a9143d9df85b2c05f0c95347e1738034e0653cd6126988ac",
                    "reqSigs": 1,
                    "type": "pubkeyhash",
                    "addresses": [
                    "mm8kkiLVQfLtLGJk52KX57SUpjXxvJ7kop"
                    ]
                }
            },
            {
                "value": 0,
                "n": 1,
                "scriptPubKey": {
                    "asm": "OP_RETURN 080112010a1803",
                    "hex": "6a07080112010a1803",
                    "type": "nulldata"
                }
            },
            {
                "value": 0,
                "n": 2,
                "scriptPubKey": {
                    "asm": "OP_DUP OP_HASH160 5f64e161b433fb843de5e19411e2a02136cda453 OP_EQUALVERIFY OP_CHECKSIG",
                    "hex": "76a9145f64e161b433fb843de5e19411e2a02136cda45388ac",
                    "reqSigs": 1,
                    "type": "pubkeyhash",
                    "addresses": [
                    "mpDMLa4N6hskcuJpTkcLTd4HB7Q2yF22bG"
                    ]
                }
            },
            {
                "value": 99.92,
                "n": 3,
                "scriptPubKey": {
                    "asm": "OP_DUP OP_HASH160 60f36fdcd16dfaba412b50d9a0af53fa2260b6a6 OP_EQUALVERIFY OP_CHECKSIG",
                    "hex": "76a91460f36fdcd16dfaba412b50d9a0af53fa2260b6a688ac",
                    "reqSigs": 1,
                    "type": "pubkeyhash",
                    "addresses": [
                    "mpManmQf6CT84xGE5zciktTmWmfHdErUQW"
                    ]
                }
            }
        ]

    assert isinstance(card_postprocess(card, vout), list)


def test_amount_to_exponent():

    assert isinstance(amount_to_exponent(88.99, 3), int)
    assert amount_to_exponent(88.99, 3) == 88990


def test_exponent_to_amount():

    assert isinstance(exponent_to_amount(10, 6), float)
    assert exponent_to_amount(10, 3) == 0.01



import pytest
import random
import itertools
from pytaoassets import Kutil
from pytaoassets.protocol import (CardTransfer, Deck, IssueMode,
                                   validate_card_issue_modes, DeckState)
from pytaoassets.exceptions import OverSizeOPReturn, InvalidCardIssue


def test_deck_object():
    '''test creation of deck objects'''

    deck = Deck(
        name="decky",
        number_of_decimals=2,
        issue_mode=IssueMode.MULTI.value,
        network="ppc",
        production=True,
        version=1,
        asset_specific_data="Just testing.",
    )

    assert deck.to_json() == {'id': None,
                              'asset_specific_data': 'Just testing.',
                              'issue_mode': IssueMode.MULTI.value,
                              'issue_time': None,
                              'tx_confirmations': None,
                              'issuer': '',
                              'name': 'decky',
                              'network': 'ppc',
                              'number_of_decimals': 2,
                              'production': True,
                              'version': 1,
                              'p2th_wif': None
                              }

    assert deck.metainfo_to_protobuf == b'\x08\x01\x12\x05decky\x18\x02 \x04*\rJust testing.'

    assert deck.metainfo_to_dict == {'issue_mode': IssueMode.MULTI.value,
                                     'name': 'decky',
                                     'number_of_decimals': 2,
                                     'version': 1,
                                     'asset_specific_data': 'Just testing.',
                                    }


def test_oversize_deck_object():
    '''test reaction to oversize deck object'''

    with pytest.raises(OverSizeOPReturn):

        deck = Deck(
            name="oversize_super_mega_deck_1119",
            number_of_decimals=18,
            issue_mode=IssueMode.SUBSCRIPTION.value,
            network="ppc",
            production=True,
            version=1,
            asset_specific_data="""
                                Lorem ipsum dolor sit amet, consectetur adipiscing elit.
                                Phasellus tempor augue at tincidunt convallis.
                                Duis laoreet ultrices augue a elementum.
                                Mauris vel dolor at mi interdum pellentesque non eget justo.
                                Maecenas tortor turpis, sodales hendrerit purus nec,
                                mattis molestie libero. Aliquam sagittis ac urna vel ultricies.
                                Fusce dictum facilisis arcu, nec ullamcorper turpis suscipit eu.
                                Etiam ac leo et turpis bibendum pharetra non consectetur quam.
                                Nullam eleifend ligula in mollis auctor.
                                Vestibulum aliquam, nulla vitae venenatis vehicula, quam nibh cursus velit,
                                vel sodales dolor magna rhoncus ipsum.
                                Nulla ante nisl, condimentum ut metus et,
                                egestas mattis velit. In lorem erat, mattis sed ex nec,
                                tincidunt dictum purus. Duis nec ligula eros.""",
        )

        assert deck.metainfo_to_protobuf


def test_card_transfer_object():

    deck = Deck(
        name="decky",
        number_of_decimals=2,
        issue_mode=IssueMode.MULTI.value,
        network="ppc",
        production=True,
        version=1,
        asset_specific_data="Just testing.",
    )

    card_transfer = CardTransfer(
        deck=deck,
        receiver=["PDZ9MPBPPjtT6qdJm98PhLVY9gNtFUoSLT"],
        amount=[1],
        version=1,
    )

    assert card_transfer.metainfo_to_protobuf == b'\x08\x01\x12\x01\x01\x18\x02'

    assert card_transfer.to_json() == {'amount': [1],
                                       'network': deck.network,
                                       'asset_specific_data': None,
                                       'blockhash': '',
                                       'blocknum': 0,
                                       'blockseq': 0,
                                       'cardseq': 0,
                                       'tx_confirmations': 0,
                                       'deck_id': None,
                                       'number_of_decimals': 2,
                                       'receiver': ['PDZ9MPBPPjtT6qdJm98PhLVY9gNtFUoSLT'],
                                       'sender': None,
                                       'timestamp': 0,
                                       'txid': None,
                                       'type': 'CardTransfer',
                                       'version': 1,
                                       'deck_p2th': None
                                       }


def test_oversize_card_object():

    deck = Deck(
        name="decky",
        number_of_decimals=12,
        issue_mode=IssueMode.MULTI.value,
        network="ppc",
        production=True,
        version=1,
        asset_specific_data="Just testing.",
    )

    with pytest.raises(OverSizeOPReturn):

        card_transfer = CardTransfer(
            deck=deck,
            receiver=[Kutil(network='ppc').address for i in range(80)],
            amount=[random.randint(20, 12000) for i in range(80)],
            version=1,
        )

        assert card_transfer.metainfo_to_protobuf


def test_card_types_issue():

    deck = Deck(
        name="decky",
        issuer='muMpqVjUDq5voY9WnxvFb9sFvZm8wwKihu',
        number_of_decimals=1,
        issue_mode=IssueMode.MULTI.value,
        network="tppc",
        production=True,
        version=1,
    )

    card_transfer = CardTransfer(
        deck=deck,
        sender=deck.issuer,
        receiver=['mmsiUudS9W5xLoWeA44JmKa28cioFg7Yzx'],
        amount=[1],
        version=1,
    )

    assert card_transfer.type == 'CardIssue'


def test_card_types_invalid():

    deck = Deck(
        name="decky",
        issuer='muMpqVjUDq5voY9WnxvFb9sFvZm8wwKihu',
        number_of_decimals=1,
        issue_mode=IssueMode.MONO.value,
        network="tppc",
        production=True,
        version=1,
    )

    with pytest.raises(InvalidCardIssue):
        card_transfer = CardTransfer(
            deck=deck,
            sender=deck.issuer,
            receiver=[deck.issuer],
            amount=[1],
            version=1,
        )


def test_card_types_transfer():

    deck = Deck(
        name="decky",
        issuer='muMpqVjUDq5voY9WnxvFb9sFvZm8wwKihu',
        number_of_decimals=1,
        issue_mode=IssueMode.MONO.value,
        network="tppc",
        production=True,
        version=1,
    )

    card_transfer = CardTransfer(
        deck=deck,
        sender='mxrr8ALSs9fmHEszs5y1w5tRsDv9r7M2bK',
        receiver=['mxrr8ALSs9fmHEszs5y1w5tRsDv9r7M2bK'],
        amount=[1],
        version=1,
    )

    assert card_transfer.type == 'CardTransfer'


@pytest.mark.parametrize("combo", [IssueMode.ONCE, IssueMode.MULTI, IssueMode.MONO])
def test_issue_mode_combos(combo):

    base_issue_mode = IssueMode.CUSTOM

    if combo == IssueMode.ONCE:
        assert base_issue_mode.value + combo.value == 3

    if combo == IssueMode.MULTI:
        assert base_issue_mode.value + combo.value == 5

    if combo == IssueMode.MONO:
        assert base_issue_mode.value + combo.value == 9


def few_random_cards(deck: Deck, n: int, card_type: str='random',
                     amount: int=None) -> list:
    '''returns <n> randomly generated cards'''

    types = ['CardIssue', 'CardBurn', 'CardTransfer']

    cards = [CardTransfer(
        deck=deck,
        receiver=[Kutil(network='tppc').address],
        amount=[random.randint(1, 100)],
        ) for i in range(n)]

    if card_type == 'transfer':
        for i in cards:
            i.__setattr__('type', 'CardTransfer')

    if card_type == 'random':
        for i in cards:
            i.__setattr__('type', random.choice(types))

    if card_type == 'issue':
        for i in cards:
            i.__setattr__('type', 'CardIssue')

    if card_type == 'burn':
        for i in cards:
            i.__setattr__('type', 'CardBurn')

    if amount:  # if there is strict requirement for amount to be <int>
        for i in cards:
            i.amount = [amount]

    return cards


def test_validate_multi_card_issue_mode():
    '''test card filtering against MULTI deck'''

    deck = Deck(
        name="decky",
        number_of_decimals=2,
        issue_mode=IssueMode.MULTI.value,
        network="tppc",
        production=True,
        version=1,
        )

    cards = few_random_cards(deck, 4, 'issue')

    assert len(validate_card_issue_modes(deck.issue_mode, cards)) == 4


def test_validate_once_card_issue_mode():
    '''test card filtering against ONCE deck'''

    deck = Deck(
        name="decky",
        number_of_decimals=2,
        issue_mode=IssueMode.ONCE.value,
        network="tppc",
        production=True,
        version=1,
        )

    cards = few_random_cards(deck, 8, 'issue')

    assert len(validate_card_issue_modes(deck.issue_mode, cards)) == 1


def test_validate_none_card_issue_mode():
    '''test card filtering against None deck'''

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=IssueMode.NONE.value,
        network="tppc",
        production=True,
        version=1,
        )

    cards = few_random_cards(deck, 8, 'issue')

    assert len(validate_card_issue_modes(deck.issue_mode, cards)) == 0


def test_validate_unflushable_card_issue_mode():
    '''test card filtering against None deck'''

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=IssueMode.UNFLUSHABLE.value,
        network="tppc",
        production=True,
        version=1,
        )

    cards_issues = few_random_cards(deck, 8, 'issue')
    random_cards = few_random_cards(deck, 16, 'transfer')

    assert len(validate_card_issue_modes(deck.issue_mode, cards_issues + random_cards)) == 8


def test_validate_mono_card_issue_mode():
    '''test card filtering against MONO deck'''

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=IssueMode.MONO.value,
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 10, 'issue')
    other = few_random_cards(deck, 9, 'transfer')

    assert len(validate_card_issue_modes(deck.issue_mode, issues + other)) == 19


def test_validate_singlet_card_issue_mode():
    '''test card filtering against SINGLET deck'''

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=IssueMode.SINGLET.value,
        network="tppc",
        production=True,
        version=1,
        )

    first_issue = few_random_cards(deck, 10, 'issue', 1)  # first 10, with amount=1
    second_issue = few_random_cards(deck, 5, 'issue')  # with random amounts

    assert len(validate_card_issue_modes(deck.issue_mode, first_issue + second_issue)) == 1


def test_validate_subscription_card_issue_mode():
    '''test card filtering against SUBSCRIPTION deck'''

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=IssueMode.SUBSCRIPTION.value,
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 10, 'issue')
    other = few_random_cards(deck, 9, 'transfer', 1)

    assert len(validate_card_issue_modes(deck.issue_mode, issues + other)) == 10


def test_validate_3combo_card_issue_mode():
    '''combo ONCE [2] and CUSTOM [1]'''

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=3,
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 8, 'issue')
    other = few_random_cards(deck, 20, 'transfer')

    assert len(validate_card_issue_modes(deck.issue_mode, issues + other)) == 21


def test_validate_10combo_card_issue_mode():
    '''combo ONCE [2] and MONO [8]'''

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=10,
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 3, 'issue')
    other = few_random_cards(deck, 20, 'transfer')

    assert len(validate_card_issue_modes(
               deck.issue_mode, issues + other)) == 21


def test_validate_6combo_card_issue_mode():
    '''combo ONCE [2] and MULTI [4]'''

    deck = Deck(
        name="decky",
        number_of_decimals=1,
        issue_mode=6,
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 3, 'issue',)
    other = few_random_cards(deck, 2, 'transfer')

    assert len(validate_card_issue_modes(
               deck.issue_mode, issues + other)) == 3


def test_validate_28combo_card_issue_mode():
    '''combo MULTI [4], MONO [8] and UNFLUSHABLE [10]'''

    mode = IssueMode.MULTI.value | IssueMode.MONO.value | IssueMode.UNFLUSHABLE.value  # 28

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=mode,
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 10, 'issue')
    other = few_random_cards(deck, 2, 'transfer', 1)

    assert len(validate_card_issue_modes(
               deck.issue_mode, issues + other)) == 10


def test_validate_13combo_card_issue_mode():
    '''combo CUSTOM [1], MULTI [4] and MONO [8]'''

    mode = IssueMode.CUSTOM.value | IssueMode.MULTI.value | IssueMode.MONO.value  # 13

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=mode,
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 10, 'issue')
    other = few_random_cards(deck, 2, 'transfer', 1)

    assert len(validate_card_issue_modes(
               deck.issue_mode, issues + other)) == 12


@pytest.mark.parametrize("combo", list(
                         itertools.combinations(
                             [0, 1, 2, 4, 8, 16, 52, 10], 2))
                         )
def test_validate_wild_two_way_combos(combo):

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=combo[0] + combo[1],
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 5, 'issue')
    other = few_random_cards(deck, 15, 'transfer')

    assert isinstance(validate_card_issue_modes(
                      deck.issue_mode, issues + other), list)


@pytest.mark.parametrize("combo", list(
                         itertools.combinations(
                             [0, 1, 2, 4, 8, 16, 52, 10], 3))
                         )
def test_validate_wild_three_way_combos(combo):

    deck = Deck(
        name="decky",
        number_of_decimals=0,
        issue_mode=combo[0] + combo[1] + combo[2],
        network="tppc",
        production=True,
        version=1,
        )

    issues = few_random_cards(deck, 5, 'issue')
    other = few_random_cards(deck, 15, 'transfer')

    assert isinstance(validate_card_issue_modes(
                      deck.issue_mode, issues + other), list)


def test_deck_state():
    '''test DeckState calculations'''

    deck = Deck(
        name="my_test_deck",
        number_of_decimals=0,
        issue_mode=4,  # MULTI
        network="tppc",
        production=True,
        version=1,
        issuer='msnHPXDWuJhRBPVNQnwXdKvEMQHLr9z1P5'
        )

    receiver_roster = ['mzsMJgqVABFhrEGrqKH7qURhmxESx4K8Ti',
                       'mmsiUudS9W5xLoWeA44JmKa28cioFg7Yzx',
                       'muMpqVjUDq5voY9WnxvFb9sFvZm8wwKihu',
                       'mxrr8ALSs9fmHEszs5y1w5tRsDv9r7M2bK']
    amounts = [10, 20, 30, 40]

    card_issues = [CardTransfer(deck=deck,
                                receiver=[r],
                                amount=[a],
                                sender=deck.issuer,
                                blockseq=0,
                                blocknum=1,
                                blockhash='d9ec32b461d80b6a549a09f5ddd550f6e2fa9021f8efe4fd7413be6c471c0b56',
                                txid='fe8f88c2a3a700a664f9547cb9c48466f900553d0a6bdb504ad52340ef00c9a0',
                                cardseq=amounts.index(a)
                                ) for r, a in zip(receiver_roster, amounts)]

    transfers = []  # list of card transfers

    # first member of the roster sends it's 10 cards to third member of the roster
    transfers.append(CardTransfer(deck=deck,
                                  sender=receiver_roster[0],
                                  receiver=[receiver_roster[2]],
                                  amount=[amounts[0]],
                                  blockhash='c5a03576178843eb5a1f1e6b878678f2c7d47b6f561fe06059e0518645b8e50e',
                                  blocknum=2,
                                  blockseq=1,
                                  cardseq=0,
                                  txid='08c886a43ce9f95a5673bc95374259b0f9eca9de1e5fb9bb7aa7826834820133',
                                  type='CardTransfer'
                                  ))

    # second member of the roster burns it's 20 cards, he calls it a scam too
    transfers.append(CardTransfer(deck=deck,
                                  sender=receiver_roster[1],
                                  receiver=[deck.issuer],  # burn
                                  amount=[amounts[1]],
                                  blockseq=1,
                                  blockhash='d6cecad875b05e9b34cb05680de0bee4f5d69ba83df23a6b6a14d1090dc992e3',
                                  cardseq=0,
                                  blocknum=3,
                                  txid='b27161ba476d29c2255d097aaa4e236752b9891a46d1fdb88f5225ee677b976e',
                                  type='CardBurn'
                                  ))

    # third member of the roster sends out it's cards to r[0] and r[3]
    transfers.append(CardTransfer(deck=deck,
                                  sender=receiver_roster[2],
                                  receiver=[receiver_roster[0]],
                                  amount=[10],
                                  blockseq=1,
                                  blocknum=5,
                                  blockhash='d638dc2d60623d16cb6b39fc165a6e7514a28c426b02db32058b87fada1cabdb',
                                  cardseq=0,
                                  txid='ebe36158ca3f364910f8a1c0f9b1b2696bed4522f84551bdb42ffd57360ce232',
                                  type='CardTransfer'
                                  ))

    transfers.append(CardTransfer(deck=deck,
                                  sender=receiver_roster[2],
                                  receiver=[receiver_roster[3]],
                                  amount=[20],
                                  blockseq=1,
                                  blocknum=5,
                                  blockhash='d638dc2d60623d16cb6b39fc165a6e7514a28c426b02db32058b87fada1cabdb',
                                  txid='ebe36158ca3f364910f8a1c0f9b1b2696bed4522f84551bdb42ffd57360ce232',
                                  type='CardTransfer',
                                  cardseq=1
                                  ))

    # fourth member of the roster is sending 10 of it's cards to second member
    transfers.append(CardTransfer(deck=deck,
                                  sender=receiver_roster[3],
                                  receiver=[receiver_roster[1]],
                                  amount=[10],
                                  blockseq=1,
                                  blocknum=200,
                                  blockhash='2896066f76f0c0f609ee0e92d195d0eb48891b91f90fa4c9a51381e9f9510b7a',
                                  txid='764afbfe6b3cecd3be8161fef363a08b8b14e7c631b4b7fbbc8edbc1475ab0fe',
                                  type='CardTransfer',
                                  cardseq=0
                                  ))

    state = DeckState(card_issues + transfers)

    assert len(state.cards) == 9
    assert len(list(state.processed_burns)) == 1
    assert len(list(state.processed_issues)) == 4
    assert len(list(state.processed_transfers)) == 4
    assert state.checksum

    assert state.balances[receiver_roster[0]] == 10
    assert state.balances[receiver_roster[1]] == 10
    assert state.balances[receiver_roster[2]] == 10
    assert state.balances[receiver_roster[3]] == 50




'''contains main protocol logic like assembly of proof-of-timeline and parsing deck info'''

import concurrent.futures
from typing import Iterator, Generator, Optional

from pytaoassets.protocol import (Deck,
                                   CardBundle,
                                   CardTransfer,
                                   validate_card_issue_modes
                                   )

from pytaoassets.provider import Provider, RpcNode

from pytaoassets.pautils import (deck_parser,
                                  find_deck_spawns,
                                  bundle_parser,
                                  tx_serialization_order,
                                  find_tx_sender
                                  )

from pytaoassets.exceptions import EmptyP2THDirectory

from pytaoassets.transactions import (nulldata_script, tx_output,
                                       p2pkh_script,
                                       make_raw_transaction,
                                       Transaction,
                                       Locktime)

from pytaoassets.pa_constants import param_query
from pytaoassets.networks import net_query
from decimal import Decimal


def find_all_valid_decks(provider: Provider, deck_version: int,
                         prod: bool=True) -> Generator:
    '''
    Scan the blockchain for PeerAssets decks, returns list of deck objects.
    : provider - provider instance
    : version - deck protocol version (0, 1, 2, ...)
    : test True/False - test or production P2TH
    '''

    pa_params = param_query(provider.network)

    if prod:
        p2th = pa_params.P2TH_addr
    else:
        p2th = pa_params.test_P2TH_addr

    if isinstance(provider, RpcNode):
        deck_spawns = (provider.getrawtransaction(i, 1)
                       for i in find_deck_spawns(provider))

    else:
        try:
            deck_spawns = (provider.getrawtransaction(i, 1) for i in
                           provider.listtransactions(p2th))
        except TypeError as err:  # it will except if no transactions are found on this P2TH
            raise EmptyP2THDirectory(err)

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as th:
        for result in th.map(deck_parser, ((provider, rawtx, deck_version, p2th) for rawtx in deck_spawns)):
            if result:
                yield result


def find_deck(provider: Provider, key: str, version: int, prod: bool=True) -> Optional[Deck]:
    '''Find specific deck by deck id.'''

    pa_params = param_query(provider.network)
    if prod:
        p2th = pa_params.P2TH_addr
    else:
        p2th = pa_params.test_P2TH_addr

    rawtx = provider.getrawtransaction(key, 1)
    deck = deck_parser((provider, rawtx, 1, p2th))

    return deck


def deck_spawn(provider: Provider, deck: Deck, inputs: dict,
               change_address: str, locktime: int=0) -> Transaction:

    '''Creates Deck spawn raw transaction.

       : key - Kutil object which we'll use to sign the tx
       : deck - Deck object
       : card - CardTransfer object
       : inputs - utxos (has to be owned by deck issuer)
       : change_address - address to send the change to
       : locktime - tx locked until block n=int
    '''

    network_params = net_query(deck.network)
    pa_params = param_query(deck.network)

    if deck.production:
        p2th_addr = pa_params.P2TH_addr
    else:
        p2th_addr = pa_params.test_P2TH_addr

    #  first round of txn making is done by presuming minimal fee
    change_sum = Decimal(inputs['total'] - network_params.min_tx_fee - pa_params.P2TH_fee)

    txouts = [
        tx_output(network=deck.network, value=pa_params.P2TH_fee,
                  n=0, script=p2pkh_script(address=p2th_addr,
                                           network=deck.network)),  # p2th

        tx_output(network=deck.network, value=Decimal(0),
                  n=1, script=nulldata_script(deck.metainfo_to_protobuf)),  # op_return

        tx_output(network=deck.network, value=change_sum,
                  n=2, script=p2pkh_script(address=change_address,
                                           network=deck.network))  # change
              ]

    unsigned_tx = make_raw_transaction(network=deck.network,
                                       inputs=inputs['utxos'],
                                       outputs=txouts,
                                       locktime=Locktime(locktime)
                                       )
    return unsigned_tx


def deck_transfer(provider: Provider, deck: Deck,
                  inputs: list, change_address: str) -> Transaction:
    '''
    The deck transfer transaction is a special case of the deck spawn transaction.
    Instead of registering a new asset, the deck transfer transaction transfers ownership from vin[1] to vin[0],
    meaning that both parties are required to sign the transfer transaction for it to be accepted in the blockchain.
    '''
    raise NotImplementedError


def find_card_bundles(provider: Provider, deck: Deck) -> Optional[Iterator]:
    '''each blockchain transaction can contain multiple cards,
       wrapped in bundles. This method finds and returns those bundles.'''

    if isinstance(provider, RpcNode):
        if deck.id is None:
            raise Exception("deck.id required to listtransactions")

        batch_data = [('getrawtransaction', [i["txid"], 1]) for
                      i in provider.listtransactions(deck.id)]
        result = provider.batch(batch_data)

        if result is not None:
            raw_txns = [i['result'] for i in result if result]

        else:
            raise EmptyP2THDirectory({'error': 'No cards found on this deck.'})

    else:
        if deck.p2th_address is None:
            raise Exception("deck.p2th_address required to listtransactions")

        try:
            raw_txns = (provider.getrawtransaction(i, 1) for i in
                        provider.listtransactions(deck.p2th_address))
        except TypeError:
            raise EmptyP2THDirectory({'error': 'No cards found on this deck.'})

    return (CardBundle(deck=deck,
                       blockhash=i['blockhash'],
                       txid=i['txid'],
                       timestamp=i['time'],
                       blockseq=tx_serialization_order(provider, i["blockhash"],
                                                       i["txid"]),
                       blocknum=provider.getblock(i["blockhash"])["height"],
                       sender=find_tx_sender(provider, i),
                       vouts=i['vout'],
                       tx_confirmations=i['confirmations']
                       ) for i in raw_txns)


def get_card_bundles(provider: Provider, deck: Deck) -> Generator:
    '''get all <deck> card bundles, if they match the protocol'''

    bundles = find_card_bundles(provider, deck)

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as th:
        for result in th.map(bundle_parser, bundles):
            if result:
                yield result


def find_all_valid_cards(provider: Provider, deck: Deck) -> Generator:
    '''find all the valid cards on this deck,
       filtering out cards which don't play nice with deck issue mode'''

    # validate_card_issue_modes must recieve a full list of cards, not batches
    unfiltered = (card for batch in get_card_bundles(provider, deck) for card in batch)

    for card in validate_card_issue_modes(deck.issue_mode, list(unfiltered)):
        yield card


def card_transfer(provider: Provider, card: CardTransfer, inputs: dict,
                  change_address: str, locktime: int=0) -> Transaction:

    '''Prepare the CardTransfer Transaction object

       : card - CardTransfer object
       : inputs - utxos (has to be owned by deck issuer)
       : change_address - address to send the change to
       : locktime - tx locked until block n=int
       '''

    network_params = net_query(provider.network)
    pa_params = param_query(provider.network)

    if card.deck_p2th is None:
        raise Exception("card.deck_p2th required for tx_output")

    outs = [
        tx_output(network=provider.network,
                  value=pa_params.P2TH_fee,
                  n=0, script=p2pkh_script(address=card.deck_p2th,
                                           network=provider.network)),  # deck p2th
        tx_output(network=provider.network,
                  value=Decimal(0), n=1,
                  script=nulldata_script(card.metainfo_to_protobuf))  # op_return
    ]

    for addr, index in zip(card.receiver, range(len(card.receiver))):
        outs.append(   # TxOut for each receiver, index + 2 because we have two outs already
            tx_output(network=provider.network, value=Decimal(0), n=index+2,
                      script=p2pkh_script(address=addr,
                                          network=provider.network))
        )

    #  first round of txn making is done by presuming minimal fee
    change_sum = Decimal(inputs['total'] - network_params.min_tx_fee - pa_params.P2TH_fee)

    outs.append(
        tx_output(network=provider.network,
                  value=change_sum, n=len(outs)+1,
                  script=p2pkh_script(address=change_address,
                                      network=provider.network))
        )

    unsigned_tx = make_raw_transaction(network=provider.network,
                                       inputs=inputs['utxos'],
                                       outputs=outs,
                                       locktime=Locktime(locktime)
                                       )
    return unsigned_tx



import pytest

from pytaoassets.provider import Cryptoid, Explorer


@pytest.mark.parametrize("provider_cls", [Cryptoid, Explorer])
def test_validateaddress_peercoin(provider_cls):
    "Check Providers that can validate Tao addresses."

    provider = provider_cls(network='tao')

    # Peercoin P2PKH, P2SH addresses.
    assert provider.validateaddress("TqSeQRAiH4tkECbtFdR99ytE2AysUYqhEf") is True
    assert provider.validateaddress("2USHRW8UcSAAnHxWnPe4yX2BV5K3dbWss9") is True

    # Peercoin Testnet P2PKH address (these _should_ be False).
    assert provider.validateaddress("mj46gUeZgeD9ufU7Fvz2dWqaX6Nswtbpba") is False
    assert provider.validateaddress("n12h8P5LrVXozfhEQEqg8SFUmVKtphBetj") is False

    # Very much not Peercoin addresses.
    assert provider.validateaddress("1BFQfjM29kubskmaAsPjPCfHYphYvKA7Pj") is False
    assert provider.validateaddress("2NFNPUYRpDXf3YXEuVT6AdMesX4kyeyDjtp") is False





"""all things PeerAssets protocol."""

from enum import Enum
from operator import itemgetter
from typing import List, Optional, Generator, cast, Callable

from pytaoassets.kutil import Kutil
from pytaoassets.paproto_pb2 import DeckSpawn as deckspawnproto
from pytaoassets.paproto_pb2 import CardTransfer as cardtransferproto
from pytaoassets.exceptions import (RecieverAmountMismatch,
                                     OverSizeOPReturn,
                                     InvalidCardIssue)
from pytaoassets.card_parsers import parsers
from pytaoassets.networks import net_query


class IssueMode(Enum):

    NONE = 0x00
    # https://github.com/PeerAssets/rfcs/blob/master/0001-taoassets-transaction-specification.proto#L19
    # No issuance allowed.

    CUSTOM = 0x01
    # https://github.com/PeerAssets/rfcs/blob/master/0001-taoassets-transaction-specification.proto#L20
    # Custom issue mode, verified by client aware of this.

    ONCE = 0x02
    # https://github.com/PeerAssets/rfcs/blob/master/0001-taoassets-transaction-specification.proto#L21
    # A single card_issue transaction allowed.

    MULTI = 0x04
    # https://github.com/PeerAssets/rfcs/blob/master/0001-taoassets-transaction-specification.proto#L22
    # Multiple card_issue transactions allowed.

    MONO = 0x08
    # https://github.com/PeerAssets/rfcs/blob/master/0001-taoassets-transaction-specification.proto#L23
    # All card transaction amounts are equal to 1.

    UNFLUSHABLE = 0x10
    # https://github.com/PeerAssets/rfcs/blob/master/0001-taoassets-transaction-specification.proto#L24
    # The UNFLUSHABLE issue mode invalidates any card transfer transaction except for the card issue transaction.
    # Meaning that only the issuing entity is able to change the balance of a specific address.
    # To correctly calculate the balance of a PeerAssets addres a client should only consider the card transfer
    # transactions originating from the deck owner.

    SUBSCRIPTION = 0x34  # SUBSCRIPTION (34 = 20 | 4 | 10)
    # https://github.com/PeerAssets/rfcs/blob/master/0001-taoassets-transaction-specification.proto#L26
    # The SUBSCRIPTION issue mode marks an address holding tokens as subscribed for a limited timeframe. This timeframe is
    # defined by the balance of the account and the time at which the first cards of this token are received.
    # To check validity of a subscription one should take the timestamp of the first received cards and add the address' balance to it in hours.

    SINGLET = 0x0a  # SINGLET is a combination of ONCE and MONO (2 | 8)
    #  Singlet deck, one MONO card issunce allowed


class Deck:

    def __init__(self, name: str,
                 number_of_decimals: int,
                 issue_mode: int,
                 network: str,
                 production: bool,
                 version: int,
                 asset_specific_data: bytes=None,
                 issuer: str="",
                 issue_time: int=None,
                 id: str=None,
                 tx_confirmations: int=None) -> None:
        '''
        Initialize deck object, load from dictionary Deck(**dict) or initilize
        with kwargs Deck("deck", 3, "ONCE")
        '''

        self.version = version  # protocol version
        self.name = name  # deck name
        self.issue_mode = issue_mode  # deck issue mode
        self.number_of_decimals = number_of_decimals
        self.asset_specific_data = asset_specific_data  # optional metadata for the deck
        self.id = id
        self.issuer = issuer
        self.issue_time = issue_time
        self.tx_confirmations = tx_confirmations
        self.network = network
        self.production = production

    @property
    def p2th_address(self) -> Optional[str]:
        '''P2TH address of this deck'''

        if self.id:
            return Kutil(network=self.network,
                         privkey=bytearray.fromhex(self.id)).address
        else:
            return None

    @property
    def p2th_wif(self) -> Optional[str]:
        '''P2TH privkey in WIF format'''

        if self.id:
            return Kutil(network=self.network,
                         privkey=bytearray.fromhex(self.id)).wif
        else:
            return None

    @property
    def metainfo_to_protobuf(self) -> bytes:
        '''encode deck into protobuf'''

        deck = deckspawnproto()
        deck.version = self.version
        deck.name = self.name
        deck.number_of_decimals = self.number_of_decimals
        deck.issue_mode = self.issue_mode
        if self.asset_specific_data:
            if not isinstance(self.asset_specific_data, bytes):
                deck.asset_specific_data = self.asset_specific_data.encode()
            else:
                deck.asset_specific_data = self.asset_specific_data

        if deck.ByteSize() > net_query(self.network).op_return_max_bytes:
            raise OverSizeOPReturn('''
                        Metainfo size exceeds maximum of {max} bytes supported by this network.'''
                                   .format(max=net_query(self.network)
                                           .op_return_max_bytes))

        return deck.SerializeToString()

    @property
    def metainfo_to_dict(self) -> dict:
        '''encode deck into dictionary'''

        r = {
            "version": self.version,
            "name": self.name,
            "number_of_decimals": self.number_of_decimals,
            "issue_mode": self.issue_mode
        }

        if self.asset_specific_data:
            r.update({'asset_specific_data': self.asset_specific_data})

        return r

    def to_json(self) -> dict:
        '''export the Deck object to json-ready format'''

        d = self.__dict__
        d['p2th_wif'] = self.p2th_wif
        return d

    @classmethod
    def from_json(cls, json: dict):
        '''load the Deck object from json'''

        return cls(**json)

    def __str__(self) -> str:

        r = []
        for key in self.__dict__:
            r.append("{key}='{value}'".format(key=key, value=self.__dict__[key]))

        return ', '.join(r)


class CardBundle:

    '''On the low level, cards come in bundles.
    A single transaction can contain dozens of cards.
    CardBundle is a object which is pre-coursor to CardTransfer'''

    def __init__(self,
                 deck: Deck,
                 sender: str,
                 txid: str,
                 blockhash: str,
                 blocknum: int,
                 blockseq: int,
                 timestamp: int,
                 tx_confirmations: int,
                 vouts: list=[],
                 ) -> None:

        self.deck = deck
        self.txid = txid
        self.sender = sender
        self.vouts = vouts

        if blockhash:
            self.blockhash = blockhash
            self.blockseq = blockseq
            self.timestamp = timestamp
            self.blocknum = blocknum
            self.tx_confirmations = tx_confirmations
        else:
            self.blockhash = ""
            self.blockseq = 0
            self.blocknum = 0
            self.timestamp = 0
            self.tx_confirmations = 0

    def to_json(self) -> dict:
        '''export the CardBundle object to json-ready format'''

        return self.__dict__


class CardTransfer:

    def __init__(self, deck: Deck, 
                 receiver: list=[],
                 amount: List[int]=[],
                 version: int=1,
                 blockhash: str=None,
                 txid: str=None,
                 sender: str=None,
                 asset_specific_data: bytes=None,
                 number_of_decimals: int=None,
                 blockseq: int=None,
                 cardseq: int=None,
                 blocknum: int=None,
                 timestamp: int=None,
                 tx_confirmations: int=None,
                 type: str=None) -> None:

        '''CardTransfer object, used when parsing card_transfers from the blockchain
        or when sending out new card_transfer.
        It can be initialized by passing the **kwargs and it will do the parsing,
        or it can be initialized with passed arguments.
        * deck - instance of Deck object
        * receiver - list of receivers
        * amount - list of amounts to be sent, must be integer
        * version - protocol version, default 1
        * txid - transaction ID of CardTransfer
        * sender - transaction sender
        * blockhash - block ID where the tx was first included
        * blockseq - order in which tx was serialized into block
        * timestamp - unix timestamp of the block where it was first included
        * tx_confirmations - number of confirmations of the transaction
        * asset_specific_data - extra metadata
        * number_of_decimals - number of decimals for amount, inherited from Deck object
        : type: card type [CardIssue, CardTransfer, CardBurn]'''

        if not len(receiver) == len(amount):
            raise RecieverAmountMismatch

        self.version = version
        self.network = deck.network
        self.deck_id = deck.id
        self.deck_p2th = deck.p2th_address
        self.txid = txid
        self.sender = sender
        self.asset_specific_data = asset_specific_data
        if not number_of_decimals:
            self.number_of_decimals = deck.number_of_decimals
        else:
            self.number_of_decimals = number_of_decimals

        self.receiver = receiver
        self.amount = amount

        if blockhash:
            self.blockhash = blockhash
            self.blockseq = blockseq
            self.timestamp = timestamp
            self.blocknum = blocknum
            self.cardseq = cardseq
            self.tx_confirmations = tx_confirmations
        else:
            self.blockhash = ""
            self.blockseq = 0
            self.blocknum = 0
            self.timestamp = 0
            self.cardseq = 0
            self.tx_confirmations = 0

        if self.sender == deck.issuer:
            # if deck issuer is issuing cards to the deck issuing address,
            # card is burn and issue at the same time - which is invalid!
            if deck.issuer in self.receiver:
                raise InvalidCardIssue
            else:
                # card was sent from deck issuer to any random address,
                # card type is CardIssue
                self.type = "CardIssue"

        # card was sent back to issuing address
        # card type is CardBurn
        elif self.receiver[0] == deck.issuer and not self.sender == deck.issuer:
            self.type = "CardBurn"

        # issuer is anyone else,
        # card type is CardTransfer
        else:
            self.type = "CardTransfer"

        if type:
            self.type = type

    @property
    def metainfo_to_protobuf(self) -> bytes:
        '''encode card_transfer info to protobuf'''

        card = cardtransferproto()
        card.version = self.version
        card.amount.extend(self.amount)
        card.number_of_decimals = self.number_of_decimals
        if self.asset_specific_data:
            if not isinstance(self.asset_specific_data, bytes):
                card.asset_specific_data = self.asset_specific_data.encode()
            else:
                card.asset_specific_data = self.asset_specific_data

        if card.ByteSize() > net_query(self.network).op_return_max_bytes:
            raise OverSizeOPReturn('''
                        Metainfo size exceeds maximum of {max} bytes supported by this network.'''
                                   .format(max=net_query(self.network)
                                           .op_return_max_bytes))

        return card.SerializeToString()

    @property
    def metainfo_to_dict(self) -> dict:
        '''encode card into dictionary'''

        r = {
            "version": self.version,
            "amount": self.amount,
            "number_of_decimals": self.number_of_decimals
        }

        if self.asset_specific_data:
            r.update({'asset_specific_data': self.asset_specific_data})

        return r

    def to_json(self) -> dict:
        '''export the CardTransfer object to json-ready format'''

        return self.__dict__

    @classmethod
    def from_json(cls, json: dict):
        '''load the Deck object from json'''

        return cls(**json)

    def __str__(self) -> str:

        r = []
        for key in self.__dict__:
            r.append("{key}='{value}'".format(key=key, value=self.to_json()[key]))

        return ', '.join(r)


def validate_card_issue_modes(issue_mode: int, cards: list) -> list:
    """validate cards against deck_issue modes"""

    supported_mask = 63  # sum of all issue_mode values

    if not bool(issue_mode & supported_mask):
        return []  # return empty list

    for i in [1 << x for x in range(len(IssueMode))]:
        if bool(i & issue_mode):

            try:
                parser_fn = cast(
                    Callable[[list], Optional[list]],
                    parsers[IssueMode(i).name]
                )
            except ValueError:
                continue

            parsed_cards = parser_fn(cards)
            if not parsed_cards:
                return []
            cards = parsed_cards

    return cards


class DeckState:

    def __init__(self, cards: Generator) -> None:

        self.cards = cards
        self.total = 0
        self.burned = 0
        self.balances = cast(dict, {})
        self.processed_issues = set()
        self.processed_transfers = set()
        self.processed_burns = set()

        self.calc_state()
        self.checksum = not bool(self.total - sum(self.balances.values()))

    def _process(self, card: dict, ctype: str) -> bool:

        sender = card["sender"]
        receiver = card["receiver"][0]
        amount = card["amount"][0]

        if ctype != 'CardIssue':
            balance_check = sender in self.balances and self.balances[sender] >= amount

            if balance_check:
                self.balances[sender] -= amount

                if 'CardBurn' not in ctype:
                    self._append_balance(amount, receiver)

                return True

            return False

        if 'CardIssue' in ctype:
            self._append_balance(amount, receiver)
            return True

        return False

    def _append_balance(self, amount: int, receiver: str) -> None:

            try:
                self.balances[receiver] += amount
            except KeyError:
                self.balances[receiver] = amount

    def _sort_cards(self, cards: Generator) -> list:
        '''sort cards by blocknum and blockseq'''

        return sorted([card.__dict__ for card in cards],
                            key=itemgetter('blocknum', 'blockseq', 'cardseq'))

    def calc_state(self) -> None:

        for card in self._sort_cards(self.cards):

            # txid + blockseq + cardseq, as unique ID
            cid = str(card["txid"] + str(card["blockseq"]) + str(card["cardseq"]))
            ctype = card["type"]
            amount = card["amount"][0]

            if ctype == 'CardIssue' and cid not in self.processed_issues:
                validate = self._process(card, ctype)
                self.total += amount * validate  # This will set amount to 0 if validate is False
                self.processed_issues |= {cid}

            if ctype == 'CardTransfer' and cid not in self.processed_transfers:
                self._process(card, ctype)
                self.processed_transfers |= {cid}

            if ctype == 'CardBurn' and cid not in self.processed_burns:
                validate = self._process(card, ctype)

                self.total -= amount * validate
                self.burned += amount * validate
                self.processed_burns |= {cid}



import pytest

from pytaoassets.exceptions import UnsupportedNetwork
from pytaoassets.networks import net_query


def test_net_query():
    "Check that we can find NetworkParams for networks by name."

    # Use a network's long name
    net_params = net_query("tao")
    assert net_params.shortname == "xto"

    # Use a network's short name
    net_params = net_query("xto")
    assert net_params.name == "tao"

    # Try to find a network we don't know about.
    with pytest.raises(UnsupportedNetwork):
        net_query("not a network name we know of.")



from pytaoassets.kutil import Kutil
from pytaoassets.provider import *
from pytaoassets.__main__ import (deck_parser,
                                   find_all_valid_cards,
                                   find_all_valid_decks,
                                   find_deck,
                                   deck_spawn,
                                   deck_transfer,
                                   get_card_bundles,
                                   card_transfer)
from pytaoassets.protocol import Deck, CardTransfer, DeckState



import warnings
from typing import Iterable, List

from pytaoassets.kutil import Kutil
from pytaoassets.protocol import Deck
from pytaoassets.provider import Provider
from pytaoassets import pavoteproto_pb2 as pavoteproto
from hashlib import sha256
from pytaoassets import transactions
from pytaoassets.pautils import read_tx_opreturn, find_tx_sender
from pytaoassets.networks import net_query


def deck_vote_tag(deck: Deck) -> str:
    '''deck vote tag address'''

    if deck.id is None:
        raise Exception("deck.id is required")

    deck_vote_tag_privkey = sha256(unhexlify(deck.id) + b"vote_init").hexdigest()
    deck_vote_tag_address = Kutil(network=deck.network, privkey=bytearray.fromhex(deck_vote_tag_privkey))
    return deck_vote_tag_address.address


class Vote:

    def __init__(self, version: int, description: str, count_mode: str,
                 start_block: int, end_block: int, deck: Deck,
                 choices: list=[], vote_metainfo: str="", vote_id: str=None, sender: str=None) -> None:
        '''initialize vote object'''

        self.version = version
        self.description = description
        self.choices = choices
        self.count_mode = count_mode
        self.start_block = start_block  # at which block does vote start
        self.end_block = end_block  # at which block does vote end
        self.vote_id = vote_id  # vote_init txid
        self.vote_metainfo = vote_metainfo  # any extra info describing the vote
        self.sender = sender
        self.deck = deck

    @property
    def to_protobuf(self) -> str:
        '''encode vote into protobuf'''

        vote = pavoteproto.Vote()
        vote.version = self.version
        vote.description = self.description
        vote.count_mode = vote.MODE.Value(self.count_mode)
        vote.start_block = self.start_block
        vote.end_block = self.end_block
        vote.choices.extend(self.choices)

        if not isinstance(self.vote_metainfo, bytes):
            vote.vote_metainfo = self.vote_metainfo.encode()
        else:
            vote.vote_metainfo = self.vote_metainfo

        proto = vote.SerializeToString()

        if len(proto) > 80:
            warnings.warn('\nMetainfo size exceeds maximum of 80 bytes allowed by OP_RETURN.')

        return proto

    @property
    def to_dict(self) -> dict:
        '''vote info as dict'''

        return {
            "version": self.version,
            "description": self.description,
            "count_mode": self.count_mode,
            "start_block": self.start_block,
            "end_block": self.end_block,
            "choices": self.choices,
            "vote_metainfo": self.vote_metainfo
        }

    @property
    def vote_choice_address(self) -> List[str]:
        '''calculate the addresses on which the vote is casted.'''

        if self.vote_id is None:
            raise Exception("vote_id is required")

        addresses = []
        vote_init_txid = unhexlify(self.vote_id)

        for choice in self.choices:
            vote_cast_privkey = sha256(vote_init_txid + bytes(
                                    list(self.choices).index(choice))
                                    ).hexdigest()
            addresses.append(Kutil(network=self.deck.network,
                                   privkey=bytearray.fromhex(vote_cast_privkey)).address)

        return addresses


def parse_vote_info(protobuf: bytes) -> dict:
    '''decode vote init tx op_return protobuf message and validate it.'''

    vote = pavoteproto.Vote()
    vote.ParseFromString(protobuf)

    assert vote.version > 0, {"error": "Vote info incomplete, version can't be 0."}
    assert vote.start_block < vote.end_block, {"error": "vote can't end in the past."}

    return {
        "version": vote.version,
        "description": vote.description,
        "count_mode": vote.MODE.Name(vote.count_mode),
        "choices": vote.choices,
        "start_block": vote.start_block,
        "end_block": vote.end_block,
        "vote_metainfo": vote.vote_metainfo
    }


def vote_init(vote: Vote, inputs: dict, change_address: str) -> bytes:
    '''initialize vote transaction, must be signed by the deck_issuer privkey'''

    network_params = net_query(vote.deck.network)
    deck_vote_tag_address = deck_vote_tag(vote.deck)

    tx_fee = network_params.min_tx_fee  # settle for min tx fee for now

    for utxo in inputs['utxos']:
        utxo['txid'] = unhexlify(utxo['txid'])
        utxo['scriptSig'] = unhexlify(utxo['scriptSig'])

    outputs = [
        {"redeem": 0.01, "outputScript": transactions.monosig_script(deck_vote_tag_address)},
        {"redeem": 0, "outputScript": transactions.op_return_script(vote.to_protobuf)},
        {"redeem": float(inputs['total']) - float(tx_fee) - float(0.01),
         "outputScript": transactions.monosig_script(change_address)
         }]

    return transactions.make_raw_transaction(inputs['utxos'], outputs)


def find_vote_inits(provider: Provider, deck: Deck) -> Iterable[Vote]:
    '''find vote_inits on this deck'''

    vote_ints = provider.listtransactions(deck_vote_tag(deck))

    for txid in vote_ints:
        try:
            raw_vote = provider.getrawtransaction(txid)
            vote = parse_vote_info(read_tx_opreturn(raw_vote))
            vote["vote_id"] = txid
            vote["sender"] = find_tx_sender(provider, raw_vote)
            vote["deck"] = deck
            yield Vote(**vote)
        except AssertionError:
            pass


def vote_cast(vote: Vote, choice_index: int, inputs: dict,
              change_address: str) -> bytes:
    '''vote cast transaction'''

    network_params = net_query(vote.deck.network)
    vote_cast_addr = vote.vote_choice_address[choice_index]

    tx_fee = network_params.min_tx_fee  # settle for min tx fee for now

    for utxo in inputs['utxos']:
        utxo['txid'] = unhexlify(utxo['txid'])
        utxo['scriptSig'] = unhexlify(utxo['scriptSig'])

    outputs = [
        {"redeem": 0.01, "outputScript": transactions.monosig_script(vote_cast_addr)},
        {"redeem": float(inputs['total']) - float(tx_fee) - float(0.01),
         "outputScript": transactions.monosig_script(change_address)
         }]

    return transactions.make_raw_transaction(inputs['utxos'], outputs)


class VoteCast:
    '''vote cast object, internal represtentation of the vote_cast transaction'''

    def __init__(self, vote: Vote, sender: str, blocknum: int,
                 confirmations: int, timestamp: int) -> None:
        self.vote = vote
        self.sender = sender
        self.blocknum = blocknum
        self.confirmations = confirmations
        self.timestamp = timestamp

    @property
    def is_valid(self) -> bool:
        '''check if VoteCast is valid'''

        if not (self.blocknum >= self.vote.start_block and
                self.blocknum <= self.vote.end_block):
            return False

        if not self.confirmations >= 6:
            return False

        return True


def find_vote_casts(provider: Provider, vote: Vote, choice_index: int) -> Iterable[VoteCast]:
    '''find and verify vote_casts on this vote_choice_address'''

    vote_casts = provider.listtransactions(vote.vote_choice_address[choice_index])
    for tx in vote_casts:
        raw_tx = provider.getrawtransaction(tx, 1)

        sender = find_tx_sender(provider, raw_tx)
        confirmations = raw_tx["confirmations"]
        blocknum = provider.getblock(raw_tx["blockhash"])["height"]
        yield VoteCast(vote, sender, blocknum, confirmations, raw_tx["blocktime"])



'''transaction assembly/dissasembly'''

from decimal import Decimal
from math import ceil
from time import time

from taopy.structs.address import Address
from taopy.structs.script import (
    NulldataScript,
    P2pkhScript,
    ScriptSig,
    StackData,
)
from taopy.structs.transaction import (
    Locktime,
    PeercoinMutableTx,
    MutableTransaction,
    Transaction,
    TxIn,
    TxOut,
)

from pytaoassets.kutil import Kutil
from pytaoassets.networks import net_query
from pytaoassets.provider import Provider


def calculate_tx_fee(tx_size: int) -> Decimal:
    '''return tx fee from tx size in bytes'''

    min_fee = Decimal(0.01)  # minimum

    return Decimal(ceil(tx_size / 1000) * min_fee)


def nulldata_script(data: bytes) -> NulldataScript:
    '''create nulldata (OP_return) script'''

    stack = StackData.from_bytes(data)
    return NulldataScript(stack)


def p2pkh_script(network: str, address: str) -> P2pkhScript:
    '''create pay-to-key-hash (P2PKH) script'''

    network_params = net_query(network)

    addr = Address.from_string(network=network_params,
                               string=address)

    return P2pkhScript(addr)


def tx_output(network: str, value: Decimal, n: int,
              script: ScriptSig) -> TxOut:
    '''create TxOut object'''

    network_params = net_query(network)

    return TxOut(network=network_params,
                 value=int(value * network_params.to_unit),
                 n=n, script_pubkey=script)


def make_raw_transaction(
    network: str,
    inputs: list,
    outputs: list,
    locktime: Locktime,
    timestamp: int=int(time()),
    version: int=1,
) -> MutableTransaction:
    '''create raw transaction'''

    network_params = net_query(network)

    if network_params.name.startswith("peercoin"):
        return PeercoinMutableTx(
            version=version,
            timestamp=timestamp,
            ins=inputs,
            outs=outputs,
            locktime=locktime,
            network=network_params,
        )

    return MutableTransaction(
        version=version,
        ins=inputs,
        outs=outputs,
        locktime=locktime,
        network=network_params,
    )


def find_parent_outputs(provider: Provider, utxo: TxIn) -> TxOut:
    '''due to design of the taopy library, TxIn object must be converted to TxOut object before signing'''

    network_params = net_query(provider.network)
    index = utxo.txout  # utxo index
    return TxOut.from_json(provider.getrawtransaction(utxo.txid,
                           1)['vout'][index],
                           network=network_params)


def sign_transaction(provider: Provider, unsigned: MutableTransaction,
                     key: Kutil) -> Transaction:
    '''sign transaction with Kutil'''

    parent_outputs = [find_parent_outputs(provider, i) for i in unsigned.ins]
    return key.sign_transaction(parent_outputs, unsigned)



'''various protocol constants'''

from collections import namedtuple
from decimal import Decimal

from pytaoassets.exceptions import UnsupportedNetwork


PAParams = namedtuple('PAParams', [
    'network_name',
    'network_shortname',
    'P2TH_wif',
    'P2TH_addr',
    'test_P2TH_wif',
    'test_P2TH_addr',
    'P2TH_fee',
])

params = (

    ## Tao mainnet
    PAParams("tao", "xto", "CK4NSWrSEAmDC6YqzDWq7niK2EzZ1ufvvLLczodmsCtEj7v5sB1x",
             "TpeCntDjXS2E3rJmJDpuqkhN1v93yzzVQx", "",
             "", Decimal(0.0002)),

    PAParams("peercoin", "ppc", "U624wXL6iT7XZ9qeHsrtPGEiU78V1YxDfwq75Mymd61Ch56w47KE",
             "PAprodbYvZqf4vjhef49aThB9rSZRxXsM6", "UAbxMGQQKmfZCwKXAhUQg3MZNXcnSqG5Z6wAJMLNVUAcyJ5yYxLP",
             "PAtesth4QreCwMzXJjYHBcCVKbC4wjbYKP", Decimal(0.01)),

    ## PPC testnet
    PAParams("peercoin-testnet", "tppc", "cTJVuFKuupqVjaQCFLtsJfG8NyEyHZ3vjCdistzitsD2ZapvwYZH",
             "miHhMLaMWubq4Wx6SdTEqZcUHEGp8RKMZt", "cQToBYwzrB3yHr8h7PchBobM3zbrdGKj2LtXqg7NQLuxsHeKJtRL",
             "mvfR2sSxAfmDaGgPcmdsTwPqzS6R9nM5Bo", Decimal(0.01))
)


def param_query(name: str) -> PAParams:
    '''Find the PAParams for a network by its long or short name. Raises
    UnsupportedNetwork if no PAParams is found.
    '''

    for pa_params in params:
        if name in (pa_params.network_name, pa_params.network_shortname,):
            return pa_params

    raise UnsupportedNetwork



from decimal import Decimal, getcontext
from http.client import HTTPResponse
import json
from operator import itemgetter
from typing import Union, cast
from urllib.request import Request, urlopen

from taopy.structs.transaction import TxIn, Sequence, ScriptSig

from pytaoassets.exceptions import InsufficientFunds
from pytaoassets.provider.common import Provider


class Cryptoid(Provider):

    '''API wrapper for http://chainz.cryptoid.info blockexplorer.'''

    api_key = '7547f94398e3'
    api_url_fmt = 'https://chainz.cryptoid.info/{net}/api.dws'
    explorer_url = 'https://chainz.cryptoid.info/explorer/'

    def __init__(self, network: str) -> None:
        """
        : network = peercoin [ppc], peercoin-testnet [tppc] ...
        """

        self.net = self._netname(network)['short']
        self.api_url = self.api_url_fmt.format(net=self.format_name(self.net))
        if 'ppc' in self.net:
            getcontext().prec = 6  # set to six decimals if it's Peercoin

    @staticmethod
    def format_name(net: str) -> str:
        '''take care of specifics of cryptoid naming system'''

        if net.startswith('t') or 'testnet' in net:
            net = net[1:] + '-test'
        else:
            net = net

        return net

    @staticmethod
    def get_url(url: str) -> Union[dict, int, float, str]:
        '''Perform a GET request for the url and return a dictionary parsed from
        the JSON response.'''

        request = Request(url, headers={"User-Agent": "pytaoassets"})
        response = cast(HTTPResponse, urlopen(request))
        if response.status != 200:
            raise Exception(response.reason)
        return json.loads(response.read().decode())

    def api_req(self, query: str) -> dict:

        query = "?q=" + query + "&key=" + self.api_key
        return cast(dict, self.get_url(self.api_url + query))

    def getblockcount(self) -> int:

        return cast(int, self.api_req('getblockcount'))

    def getblock(self, blockhash: str) -> dict:
        '''query block using <blockhash> as key.'''

        query = 'block.raw.dws?coin={net}&hash={blockhash}'.format(
            net=self.format_name(self.net),
            blockhash=blockhash,
        )
        return cast(dict, self.get_url(self.explorer_url + query))

    def getblockhash(self, blocknum: int) -> str:
        '''get blockhash'''

        query = 'getblockhash' + '&height=' + str(blocknum)
        return cast(str, self.api_req(query))

    def getdifficulty(self) -> dict:

        pos_difficulty = cast(float, self.api_req('getdifficulty'))
        return {"proof-of-stake": pos_difficulty}

    def getbalance(self, address: str) -> Decimal:

        query = 'getbalance' + '&a=' + address
        return Decimal(cast(float, self.api_req(query)))

    def getreceivedbyaddress(self, address: str) -> Decimal:

        query = 'getreceivedbyaddress' + "&a=" + address
        return Decimal(cast(float, self.api_req(query)))

    def listunspent(self, address: str) -> list:

        query = 'unspent' + "&active=" + address
        return cast(dict, self.api_req(query))['unspent_outputs']

    def select_inputs(self, address: str, amount: int) -> dict:
        '''select UTXOs'''

        utxos = []
        utxo_sum = Decimal(-0.01)  # starts from negative due to minimal fee
        for tx in sorted(self.listunspent(address=address), key=itemgetter('confirmations')):

                utxos.append(
                    TxIn(txid=tx['tx_hash'],
                         txout=tx['tx_ouput_n'],
                         sequence=Sequence.max(),
                         script_sig=ScriptSig.unhexlify(tx['script']))
                         )

                utxo_sum += Decimal(int(tx['value']) / 100000000)
                if utxo_sum >= amount:
                    return {'utxos': utxos, 'total': utxo_sum}

        if utxo_sum < amount:
            raise InsufficientFunds('Insufficient funds.')

        raise Exception("undefined behavior :.(")

    def getrawtransaction(self, txid: str, decrypt: int=0) -> dict:

        query = 'tx.raw.dws?coin={net}&id={txid}'.format(
            net=self.format_name(self.net),
            txid=txid,
        )
        if not decrypt:
            query += '&hex'
            return cast(dict, self.get_url(self.explorer_url + query))['hex']

        return cast(dict, self.get_url(self.explorer_url + query))

    def listtransactions(self, address: str) -> list:

        query = 'address.summary.dws?coin={net}&id={addr}'.format(
            net=self.format_name(self.net),
            addr=address,
        )
        response = cast(dict, self.get_url(self.explorer_url + query))
        return [tx[1].lower() for tx in response["tx"]]



import pytest
import pytaoassets as pa
from pytaoassets.__main__ import find_card_bundles
from pytaoassets.protocol import CardBundle
from taopy.structs.transaction import MutableTransaction
from pytaoassets.transactions import Transaction


@pytest.mark.parametrize("prov", [pa.Explorer, pa.Cryptoid])
def test_find_deck(prov):

    provider = prov(network="tppc")

    deck = pa.find_deck(provider, 'b6a95f94fef093ee9009b04a09ecb9cb5cba20ab6f13fe0926aeb27b8671df43', 1)

    deck.tx_confirmations = 100  # make it deterministic

    assert deck.to_json() == {'asset_specific_data': b'',
                              'id': 'b6a95f94fef093ee9009b04a09ecb9cb5cba20ab6f13fe0926aeb27b8671df43',
                              'issue_mode': 4,
                              'issue_time': 1488840533,
                              'issuer': 'msYThv5bf7KjhHT1Cj5D7Y1tofyhq9vhWM',
                              'name': 'hopium_v2',
                              'network': 'peercoin-testnet',
                              'number_of_decimals': 2,
                              'production': True,
                              'version': 1,
                              'tx_confirmations': 100,
                              'p2th_wif': 'cThmj6Qu6aTUeA5f4FoNJTsBA8K6ZjhXbZkwsqcmv94xjWiCBr5d'
                              }


def test_find_card_bundles():

    provider = pa.Explorer(network="tppc")
    deck = pa.find_deck(provider, 'adc6d888508ebfcad5c182df4ae94553bae6287735d76b8d64b3de8d29fc2b5b', 1)

    bundles = find_card_bundles(provider, deck)

    assert bundles
    assert isinstance(next(bundles), CardBundle)


@pytest.mark.parametrize("prov", [pa.Explorer, pa.Cryptoid])
def test_get_card_bundles(prov):

    provider = prov(network="tppc")

    deck = pa.find_deck(provider, 'b6a95f94fef093ee9009b04a09ecb9cb5cba20ab6f13fe0926aeb27b8671df43', 1)

    bundles = pa.get_card_bundles(provider, deck)

    assert bundles
    assert isinstance(list(next(bundles))[0], pa.CardTransfer)


@pytest.mark.parametrize("prov", [pa.Explorer, pa.Cryptoid])
def test_find_all_valid_cards(prov):

    provider = prov(network="tppc")

    deck = pa.find_deck(provider, 'b6a95f94fef093ee9009b04a09ecb9cb5cba20ab6f13fe0926aeb27b8671df43', 1)

    cards = pa.find_all_valid_cards(provider, deck)

    assert cards
    assert isinstance(next(cards), pa.CardTransfer)


def test_deck_spawn():

    provider = pa.Explorer(network='tppc')
    inputs = provider.select_inputs("mthKQHpr7zUbMvLcj8GHs33mVcf91DtN6L", 0.02)
    change_address = "mthKQHpr7zUbMvLcj8GHs33mVcf91DtN6L"
    deck = pa.Deck(name="just-testing.", number_of_decimals=1, issue_mode=1,
                   network='tppc', production=True, version=1,
                   asset_specific_data='https://talk.peercoin.net/')

    deck_spawn = pa.deck_spawn(provider, deck, inputs, change_address)

    assert isinstance(deck_spawn, MutableTransaction)


def test_card_transfer():

    provider = pa.Explorer(network='tppc')
    address = "mthKQHpr7zUbMvLcj8GHs33mVcf91DtN6L"
    inputs = provider.select_inputs(address, 0.02)
    change_address = address
    deck = pa.find_deck(provider,
                        '078f41c257642a89ade91e52fd484c141b11eda068435c0e34569a5dfcce7915',
                        1, True)
    card = pa.CardTransfer(deck=deck,
                           receiver=['n12h8P5LrVXozfhEQEqg8SFUmVKtphBetj',
                                     'n422r6tcJ5eofjsmRvF6TcBMigmGbY5P7E'],
                           amount=[1, 2]
                           )

    card_transfer = pa.card_transfer(provider, card, inputs, change_address,
                                     locktime=300000)

    assert isinstance(card_transfer, Transaction)



from setuptools import setup

setup(name='pytaoassets',
      version='0.4.8',
      description='Python implementation of the PeerAssets protocol for the Tao blockchain.',
      keywords=["blockchain", "digital assets", "protocol"],
      url='https://github.com/taoblockchain/pytaoassets',
      author='Bryce Weiner',
      author_email='bryce@tao.network',
      license='BSD',
      packages=['pytaoassets', 'pytaoassets.provider'],
      install_requires=['protobuf', 'taoassets-taopy'],
      extras_require={
        'rcp_provider': ['python-taorpc>=1.0']
                        }
      )



from decimal import Decimal
from http.client import HTTPResponse
import json
from typing import Union, cast
from urllib.request import urlopen

from taopy.structs.transaction import ScriptSig, Sequence, TxIn

from pytaoassets.exceptions import InsufficientFunds, UnsupportedNetwork
from pytaoassets.provider.common import Provider


class Explorer(Provider):

    '''API wrapper for https://explorer.peercoin.net blockexplorer.'''

    def __init__(self, network: str) -> None:
        """
        : network = peercoin [ppc], peercoin-testnet [tppc] ...
        """

        self.net = self._netname(network)['short']
        if 'ppc' not in self.net:
            raise UnsupportedNetwork('This API only supports Peercoin.')

    def api_fetch(self, command: str) -> Union[dict, int, float, str]:

        apiurl = 'https://explorer.peercoin.net/api/'
        if self.is_testnet:
            apiurl = 'https://testnet-explorer.peercoin.net/api/'

        response = cast(HTTPResponse, urlopen(apiurl + command))
        if response.status != 200:
            raise Exception(response.reason)

        r = response.read()

        try:
            return json.loads(r.decode())
        except json.decoder.JSONDecodeError:
            return r.decode()

    def ext_fetch(self, command: str) -> Union[dict, int, float, str]:

        extapiurl = 'https://explorer.peercoin.net/ext/'
        if self.is_testnet:
            extapiurl = 'https://testnet-explorer.peercoin.net/ext/'

        response = cast(HTTPResponse, urlopen(extapiurl + command))
        if response.status != 200:
            raise Exception(response.reason)

        try:
            return json.loads(response.read().decode())
        except json.decoder.JSONDecodeError:
            return response.read().decode()

    def getdifficulty(self) -> dict:
        '''Returns the current difficulty.'''

        return cast(dict, self.api_fetch('getdifficulty'))

    def getconnectioncount(self) -> int:
        '''Returns the number of connections the block explorer has to other nodes.'''

        return cast(int, self.api_fetch('getconnectioncount'))

    def getblockcount(self) -> int:
        '''Returns the current block index.'''

        return cast(int, self.api_fetch('getblockcount'))

    def getblockhash(self, index: int) -> str:
        '''Returns the hash of the block at ; index 0 is the genesis block.'''

        return cast(str, self.api_fetch('getblockhash?index=' + str(index)))

    def getblock(self, hash: str) -> dict:
        '''Returns information about the block with the given hash.'''

        return cast(dict, self.api_fetch('getblock?hash=' + hash))

    def getrawtransaction(self, txid: str, decrypt: int=0) -> dict:
        '''Returns raw transaction representation for given transaction id.
        decrypt can be set to 0(false) or 1(true).'''

        q = 'getrawtransaction?txid={txid}&decrypt={decrypt}'.format(txid=txid, decrypt=decrypt)

        return cast(dict, self.api_fetch(q))

    def getnetworkghps(self) -> float:
        '''Returns the current network hashrate. (ghash/s)'''

        return cast(float, self.api_fetch('getnetworkghps'))

    def getmoneysupply(self) -> Decimal:
        '''Returns current money supply.'''

        return Decimal(cast(float, self.ext_fetch('getmoneysupply')))

    def getdistribution(self) -> dict:
        '''Returns wealth distribution stats.'''

        return cast(dict, self.ext_fetch('getdistribution'))

    def getaddress(self, address: str) -> dict:
        '''Returns information for given address.'''

        return cast(dict, self.ext_fetch('getaddress/' + address))

    def listunspent(self, address: str) -> list:
        '''Returns unspent transactions for given address.'''

        try:
            return cast(dict, self.ext_fetch('listunspent/' + address))['unspent_outputs']
        except KeyError:
            raise InsufficientFunds('Insufficient funds.')

    def select_inputs(self, address: str, amount: int) -> dict:

        utxos = []
        utxo_sum = Decimal(-0.01)  # starts from negative due to minimal fee
        for tx in self.listunspent(address=address):

                utxos.append(
                    TxIn(txid=tx['tx_hash'],
                         txout=tx['tx_ouput_n'],
                         sequence=Sequence.max(),
                         script_sig=ScriptSig.unhexlify(tx['script']))
                         )

                utxo_sum += Decimal(tx['value'] / 10**8)
                if utxo_sum >= amount:
                    return {'utxos': utxos, 'total': utxo_sum}

        if utxo_sum < amount:
            raise InsufficientFunds('Insufficient funds.')

        raise Exception("undefined behavior :.(")

    def txinfo(self, txid: str) -> dict:
        '''Returns information about given transaction.'''

        return cast(dict, self.ext_fetch('txinfo/' + txid))

    def getbalance(self, address: str) -> Decimal:
        '''Returns current balance of given address.'''

        try:
            return Decimal(cast(float, self.ext_fetch('getbalance/' + address)))
        except TypeError:
            return Decimal(0)

    def getreceivedbyaddress(self, address: str) -> Decimal:

        return Decimal(cast(float, self.getaddress(address)['received']))

    def listtransactions(self, address: str) -> list:

        try:
            r = self.getaddress(address)['last_txs']
            return [i['addresses'] for i in r]
        except KeyError:
            return None





import pytest
import time
from decimal import Decimal

from taopy.structs.transaction import (
    Locktime,
    TxOut,
    MutableTransaction,
    PeercoinMutableTx,
    Transaction
)

from taopy.structs.script import P2pkhScript

from pytaoassets.transactions import (
    calculate_tx_fee,
    make_raw_transaction,
    p2pkh_script,
    tx_output,
    sign_transaction
)

from pytaoassets.networks import net_query
import pytaoassets as pa


@pytest.mark.parametrize("tx_size", [181, 311])
def test_calculate_transaction_fee(tx_size):

    assert round(calculate_tx_fee(tx_size), 2) == round(Decimal(0.01), 2)


@pytest.mark.parametrize("network", ['tao'])
def test_tx_output(network):

    if network == 'tao':
        addr = 'TbdCceyr7sTxTKydxtDTPwoFhdDwY8T3Tv'

    script = p2pkh_script(network, addr)

    txout = tx_output(network=network, value=Decimal(1.35),
                      n=1, script=script
                      )

    assert isinstance(txout, TxOut)


def test_nulldata_script():

    null = tx_output(network='tao',
                     value=Decimal(0), n=1,
                     script='Oh Hello.'.encode('utf-8'))

    assert isinstance(null, TxOut)


def test_p2pkh_script():

    addr = 'mvWDumZZZVD2nEC7hmsX8dMSHoGHAq5b6d'
    script = p2pkh_script('xto', addr)

    assert isinstance(script, P2pkhScript)


def test_make_raw_transaction():

    tx = make_raw_transaction("tao", [], [], Locktime(300000))
    assert isinstance(tx, MutableTransaction)


def test_sign_transaction():

    network_params = net_query('xto')

    provider = pa.Cryptoid(network='xto')
    key = pa.Kutil(network='xto',
                   privkey=bytearray.fromhex('4784326D5470603F569CF9BA257B7BAFD2A5D296AC61CEC44576FE7AD5EE0C7E')
                   )
    dest_address = pa.Kutil(network='xto').address
    unspent = provider.select_inputs(key.address, 1)

    output = tx_output(network='xto',
                       value=Decimal(0.1),
                       n=0, script=p2pkh_script(network='xto',
                                                address=dest_address)
                       )

    unsigned = TaoMutableTx(version=1,
                                 timestamp=int(time.time()),
                                 ins=unspent['utxos'],
                                 outs=[output],
                                 locktime=Locktime(0),
                                 data='',
                                 network=network_params
                                 )

    assert isinstance(sign_transaction(provider, unsigned, key), Transaction)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: 0003-taoassets-alias-poid-transaction-specification.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='0003-taoassets-alias-poid-transaction-specification.proto',
  package='',
  syntax='proto3',
  serialized_pb=_b('\n:0003-taoassets-alias-poid-transaction-specification.proto\"\xc7\x01\n\x08Identity\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x11\n\tfull_name\x18\x02 \x01(\t\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x12\n\ntwitter_id\x18\x04 \x01(\t\x12\x11\n\tgithub_id\x18\x05 \x01(\t\x12\x11\n\treddit_id\x18\x06 \x01(\t\x12\x13\n\x0b\x66reenode_id\x18\x07 \x01(\t\x12\x13\n\x0b\x66\x61\x63\x65\x62ook_id\x18\x08 \x01(\r\x12\x17\n\x0fgpg_fingerprint\x18\t \x01(\x0c\x12\x0b\n\x03var\x18\n \x01(\x0c\x62\x06proto3')
)




_IDENTITY = _descriptor.Descriptor(
  name='Identity',
  full_name='Identity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='Identity.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='full_name', full_name='Identity.full_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='email', full_name='Identity.email', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='twitter_id', full_name='Identity.twitter_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='github_id', full_name='Identity.github_id', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='reddit_id', full_name='Identity.reddit_id', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='freenode_id', full_name='Identity.freenode_id', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='facebook_id', full_name='Identity.facebook_id', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='gpg_fingerprint', full_name='Identity.gpg_fingerprint', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='var', full_name='Identity.var', index=9,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=63,
  serialized_end=262,
)

DESCRIPTOR.message_types_by_name['Identity'] = _IDENTITY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Identity = _reflection.GeneratedProtocolMessageType('Identity', (_message.Message,), dict(
  DESCRIPTOR = _IDENTITY,
  __module__ = '0003_taoassets_alias_poid_transaction_specification_pb2'
  # @@protoc_insertion_point(class_scope:Identity)
  ))
_sym_db.RegisterMessage(Identity)


# @@protoc_insertion_point(module_scope)
